from typing import Optional

from wtfis.clients.base import BaseIpEnricherClient, BaseRequestsClient
from wtfis.models.ipinfo import IpInfo, IpInfoMap


class IpInfoClient(BaseRequestsClient, BaseIpEnricherClient):
    """
    IPinfo client
    """

    baseurl = "https://ipinfo.io"

    @property
    def name(self) -> str:
        return "IPinfo"

    def _get_ipinfo(self, ip: str) -> Optional[IpInfo]:
        result = self._get(f"/{ip}/json")
        return IpInfo.model_validate(result) if result else None

    def enrich_ips(self, *ips: str) -> IpInfoMap:
        map_ = {}
        for ip in ips:
            ipinfo = self._get_ipinfo(ip)
            if ipinfo:
                map_[ipinfo.ip] = ipinfo
        return IpInfoMap.model_validate(map_)



from pathlib import Path
from typing import Optional

import pytest
from rich.console import Console, RenderableType
from rich.text import Span, Text

from wtfis.clients.abuseipdb import AbuseIpDbClient
from wtfis.clients.greynoise import GreynoiseClient
from wtfis.clients.ip2whois import Ip2WhoisClient
from wtfis.clients.ipwhois import IpWhoisClient
from wtfis.clients.shodan import ShodanClient
from wtfis.clients.urlhaus import UrlHausClient
from wtfis.clients.virustotal import VTClient
from wtfis.handlers.domain import DomainHandler
from wtfis.handlers.ip import IpAddressHandler
from wtfis.models.abuseipdb import AbuseIpDb
from wtfis.models.greynoise import GreynoiseIp
from wtfis.models.ip2location import Ip2Location
from wtfis.models.ipinfo import IpInfo
from wtfis.models.ipwhois import IpWhois
from wtfis.models.shodan import ShodanIp
from wtfis.models.urlhaus import UrlHaus


class TestTheme:
    """Expected theme values for the tests"""

    panel_title = "bold yellow"
    heading_h1 = "bold bright_green on dark_green"
    heading_h2 = "bold yellow"
    table_field = "bold bright_magenta"
    table_value = "not bold default"
    popularity_source = "bright_cyan"
    inline_stat = "cyan"
    vendor_list = "cyan"
    nameserver_list = "bright_blue"
    disclaimer = "italic white on red"
    tags = "bright_white on black"
    tags_green = "bright_green on black"
    tags_red = "bright_red on black"
    product = "orange_red1"
    port = "bright_cyan"
    transport = "cyan"
    footer = "cyan"
    info = "bold green"
    warn = "bold yellow"
    error = "bold red"
    timestamp_date = "not bold default"
    timestamp_t = "dim bright_white"
    timestamp_time = "dim default"
    timestamp_z = "dim bright_white"
    asn_org = "bright_white"
    whois_org = "bright_cyan"
    urlhaus_bl_name = "bright_cyan"
    urlhaus_bl_low = "bright_white on green"
    urlhaus_bl_med = "black on yellow"
    urlhaus_bl_high = "bright_white on red"


def open_test_data(fname: str) -> str:
    path = Path(__file__).parent.resolve() / "test_data" / fname
    with open(path) as f:
        return f.read()


def abuseipdb_get_ip(ip, pool) -> AbuseIpDb:
    """Mock replacement for AbuseIpDbClient()._get_ip()"""
    return AbuseIpDb.model_validate(pool[ip])


def greynoise_get(ip, pool) -> GreynoiseIp:
    """Mock replacement for GreynoiseClient().get_ip()"""
    return GreynoiseIp.model_validate(pool[ip])


def ip2location_get(ip, pool) -> Ip2Location:
    """Mock replacement for Ip2LocationClient().get_ip()"""
    return Ip2Location.model_validate(pool[ip])


def ipinfo_get(ip, pool) -> IpInfo:
    """Mock replacement for IpInfoClient().get_ip()"""
    return IpInfo.model_validate(pool[ip])


def ipwhois_get(ip, pool) -> IpWhois:
    """Mock replacement for IpWhoisClient().get_ipwhois()"""
    return IpWhois.model_validate(pool[ip])


def shodan_get_ip(ip, pool) -> ShodanIp:
    """Mock replacement for ShodanClient().get_ip()"""
    return ShodanIp.model_validate(pool[ip])


def urlhaus_get_host(entity, pool) -> UrlHaus:
    """Mock replacement for UrlHausClient()._get_host()"""
    return UrlHaus.model_validate(pool[entity])


def timestamp_text(ts) -> Optional[RenderableType]:
    """Standard timestamp formatting"""
    theme = TestTheme()
    return Text(
        ts,
        style=theme.timestamp_date,
        spans=[
            Span(10, 11, theme.timestamp_t),
            Span(11, 19, theme.timestamp_time),
            Span(19, 20, theme.timestamp_z),
        ],
    )


@pytest.fixture(scope="module")
def test_data():
    return open_test_data


@pytest.fixture(scope="module")
def theme():
    return TestTheme()


@pytest.fixture(scope="module")
def mock_abuseipdb_get():
    return abuseipdb_get_ip


@pytest.fixture(scope="module")
def mock_greynoise_get():
    return greynoise_get


@pytest.fixture(scope="module")
def mock_ip2location_get():
    return ip2location_get


@pytest.fixture(scope="module")
def mock_ipinfo_get():
    return ipinfo_get


@pytest.fixture(scope="module")
def mock_ipwhois_get():
    return ipwhois_get


@pytest.fixture(scope="module")
def mock_shodan_get_ip():
    return shodan_get_ip


@pytest.fixture(scope="module")
def mock_urlhaus_get():
    return urlhaus_get_host


@pytest.fixture(scope="module")
def display_timestamp():
    return timestamp_text


def generate_domain_handler(max_resolutions=3):
    return DomainHandler(
        entity="www.example[.]com",
        console=Console(),
        vt_client=VTClient("dummykey"),
        ip_geoasn_client=IpWhoisClient(),
        whois_client=Ip2WhoisClient("dummykey"),
        shodan_client=ShodanClient("dummykey"),
        greynoise_client=GreynoiseClient("dummykey"),
        abuseipdb_client=AbuseIpDbClient("dummykey"),
        urlhaus_client=UrlHausClient("dummykey"),
        max_resolutions=max_resolutions,
    )


@pytest.fixture
def domain_handler():
    return generate_domain_handler


def generate_ip_handler():
    return IpAddressHandler(
        entity="1[.]1[.]1[.]1",
        console=Console(),
        vt_client=VTClient("dummykey"),
        ip_geoasn_client=IpWhoisClient(),
        whois_client=Ip2WhoisClient("dummykey"),
        shodan_client=ShodanClient("dummykey"),
        greynoise_client=GreynoiseClient("dummykey"),
        abuseipdb_client=AbuseIpDbClient("dummykey"),
        urlhaus_client=UrlHausClient("dummykey"),
    )


@pytest.fixture
def ip_handler():
    return generate_ip_handler



import json
from unittest.mock import MagicMock, patch

import pytest
from requests.exceptions import ConnectionError

from wtfis.clients.base import requests
from wtfis.exceptions import HandlerException
from wtfis.main import fetch_data
from wtfis.models.greynoise import GreynoiseIpMap
from wtfis.models.ipwhois import IpWhoisMap
from wtfis.models.virustotal import Resolutions


class TestDomainHandler:
    def test_entity_refang(self, domain_handler):
        handler = domain_handler()
        assert handler.entity == "www.example.com"

    def test_fetch_data_1(self, domain_handler, test_data):
        """Test with max_resolutions = 3 (default)"""
        handler = domain_handler()
        handler.resolutions = Resolutions.model_validate(
            json.loads(test_data("vt_resolutions_gist.json"))
        )

        handler._fetch_vt_domain = MagicMock()
        handler._fetch_vt_resolutions = MagicMock()
        handler._fetch_geoasn = MagicMock()
        handler._fetch_whois = MagicMock()
        handler._fetch_shodan = MagicMock()
        handler._fetch_greynoise = MagicMock()
        handler._fetch_urlhaus = MagicMock()

        fetch_data(MagicMock(), handler)
        handler._fetch_vt_domain.assert_called_once()
        handler._fetch_vt_resolutions.assert_called_once()
        handler._fetch_geoasn.assert_called_once()
        handler._fetch_whois.assert_called_once()
        handler._fetch_shodan.assert_called_once()
        handler._fetch_greynoise.assert_called_once()
        handler._fetch_urlhaus.assert_called_once()

    def test_fetch_data_2(self, domain_handler):
        """Test with max_resolutions = 0"""
        handler = domain_handler(0)
        handler._fetch_vt_domain = MagicMock()
        handler._fetch_vt_resolutions = MagicMock()
        handler._fetch_geoasn = MagicMock()
        handler._fetch_whois = MagicMock()
        handler._fetch_shodan = MagicMock()
        handler._fetch_greynoise = MagicMock()
        handler._fetch_urlhaus = MagicMock()

        fetch_data(MagicMock(), handler)
        handler._fetch_vt_domain.assert_called_once()
        handler._fetch_vt_resolutions.assert_not_called()
        handler._fetch_geoasn.assert_not_called()
        handler._fetch_whois.assert_called_once()
        handler._fetch_shodan.assert_not_called()
        handler._fetch_greynoise.assert_not_called()
        handler._fetch_urlhaus.assert_called_once()

        assert handler.geoasn == IpWhoisMap.model_validate({})
        assert handler.geoasn.root == {}

    @patch.object(requests.Session, "get")
    def test_vt_http_error(self, mock_requests_get, domain_handler):
        """
        Test a requests HTTPError from the VT client. This also tests the
        common_exception_handler decorator.
        """
        handler = domain_handler()
        mock_resp = requests.models.Response()

        mock_resp.status_code = 401
        mock_requests_get.return_value = mock_resp

        # Thorough test of first _fetch_* method
        with pytest.raises(HandlerException) as e:
            handler._fetch_vt_domain()

        assert (
            e.value.args[0]
            == "Error fetching data: 401 Client Error: None for url: None"
        )
        assert e.type is HandlerException  # ruff E721

        # Extra: just make sure program exits correctly
        with pytest.raises(HandlerException) as e:
            handler._fetch_vt_resolutions()

    @patch.object(requests.Session, "get")
    def test_vt_validation_error(self, mock_requests_get, domain_handler):
        """
        Test a pydantic data model ValidationError from the VT client. This also tests
        the common_exception_handler decorator.
        """
        handler = domain_handler()
        mock_resp = requests.models.Response()

        with patch.object(mock_resp, "json") as mock_resp_json:
            mock_resp.status_code = 200
            mock_resp_json.return_value = {"intentionally": "wrong data"}
            mock_requests_get.return_value = mock_resp

            # Thorough test of first _fetch_* method
            with pytest.raises(HandlerException) as e:
                handler._fetch_vt_domain()

            assert e.value.args[0].startswith(
                "Data model validation error: 1 validation error for Domain\ndata\n"
                "  Field required [type=missing, input_value={'intentionally': "
                "'wrong data'}, input_type=dict]\n"
            )
            assert e.type is HandlerException

            # Extra: just make sure program exits correctly
            with pytest.raises(HandlerException) as e:
                handler._fetch_vt_resolutions()

    @patch.object(requests.Session, "get")
    def test_ipwhois_http_error(
        self, mock_requests_get, domain_handler, capsys, test_data
    ):
        handler = domain_handler()
        handler.resolutions = Resolutions.model_validate(
            json.loads(test_data("vt_resolutions_gist.json"))
        )

        mock_resp = requests.models.Response()

        mock_resp.status_code = 500
        mock_requests_get.return_value = mock_resp

        handler._fetch_geoasn(*["1.2.3.4", "1.2.3.5"])
        assert handler.warnings[0].startswith(
            "Could not fetch IPWhois: 500 Server Error:"
        )

        handler.print_warnings()
        capture = capsys.readouterr()
        assert capture.out.startswith(
            "WARN: Could not fetch IPWhois: 500 Server Error:"
        )

    @patch.object(requests.Session, "get")
    def test_ipwhois_validation_error(self, mock_requests_get, domain_handler):
        handler = domain_handler()
        mock_resp = requests.models.Response()

        with patch.object(mock_resp, "json") as mock_resp_json:
            mock_resp.status_code = 200
            mock_resp_json.return_value = {
                "success": True,
                "intentionally": "wrong data",
            }
            mock_requests_get.return_value = mock_resp

            with pytest.raises(HandlerException) as e:
                handler._fetch_geoasn("1.2.3.4")

            assert e.value.args[0].startswith(
                "Data model validation error: 1 validation error for IpWhois\n"
            )
            assert e.type is HandlerException

    @patch.object(requests.Session, "get")
    def test_whois_http_error(self, mock_requests_get, domain_handler, capsys):
        handler = domain_handler()
        mock_resp = requests.models.Response()

        mock_resp.status_code = 401
        mock_requests_get.return_value = mock_resp

        handler._fetch_whois()
        assert handler.warnings[0].startswith(
            "Could not fetch IP2Whois: 401 Client Error:"
        )

        handler.print_warnings()
        capture = capsys.readouterr()
        assert capture.out.startswith(
            "WARN: Could not fetch IP2Whois: 401 Client Error:"
        )

    @patch.object(requests.Session, "get")
    def test_vt_resolutions_429_error(self, mock_requests_get, domain_handler, capsys):
        """
        Test fail open behavior of VT resolution fetching when rate limited
        """
        handler = domain_handler()
        mock_resp = requests.models.Response()

        mock_resp.status_code = 429
        mock_requests_get.return_value = mock_resp

        handler._fetch_vt_resolutions()
        assert handler.warnings[0].startswith(
            "Could not fetch Virustotal resolutions: 429 Client Error:"
        )

        handler.print_warnings()
        capture = capsys.readouterr()
        assert capture.out.startswith(
            "WARN: Could not fetch Virustotal resolutions: 429 Client Error:"
        )

    @patch.object(requests.Session, "get")
    def test_whois_429_error(self, mock_requests_get, domain_handler, capsys):
        """
        Test fail open behavior of whois fetching when rate limited
        Since _fetch_whois() is in the base class, this should also cover the
        IpAddressHandler use case.
        """
        handler = domain_handler()
        mock_resp = requests.models.Response()

        mock_resp.status_code = 429
        mock_requests_get.return_value = mock_resp

        handler._fetch_whois()
        assert handler.warnings[0].startswith(
            "Could not fetch IP2Whois: 429 Client Error:"
        )

        handler.print_warnings()
        capture = capsys.readouterr()
        assert capture.out.startswith(
            "WARN: Could not fetch IP2Whois: 429 Client Error:"
        )

    @patch.object(requests.Session, "get")
    def test_greynoise_429_error(
        self, mock_requests_get, domain_handler, capsys, test_data
    ):
        """
        Test fail open behavior of Greynoise when rate limited
        """
        handler = domain_handler()
        mock_resp = requests.models.Response()

        mock_resp.status_code = 429
        mock_requests_get.return_value = mock_resp

        handler.resolutions = Resolutions.model_validate(
            json.loads(test_data("vt_resolutions_gist.json"))
        )

        handler._fetch_greynoise(*["1.2.3.4", "1.2.3.5"])
        assert handler.warnings[0].startswith(
            "Could not fetch Greynoise: 429 Client Error:"
        )

        handler.print_warnings()
        capture = capsys.readouterr()
        assert capture.out.startswith(
            "WARN: Could not fetch Greynoise: 429 Client Error:"
        )

    @patch.object(requests.Session, "get")
    def test_greynoise_404_error(self, mock_requests_get, domain_handler, test_data):
        """
        Test fail open behavior of Greynoise when no IP found (404) and no warning
        message
        """
        handler = domain_handler(3)
        mock_resp = requests.models.Response()

        handler.resolutions = Resolutions.model_validate(
            json.loads(test_data("vt_resolutions_gist.json"))
        )

        mock_resp.status_code = 404
        mock_requests_get.return_value = mock_resp

        handler._fetch_greynoise(*["1.2.3.4", "1.2.3.5"])
        assert len(handler.warnings) == 0
        assert handler.greynoise == GreynoiseIpMap.model_validate({})

    @patch.object(requests.Session, "get")
    def test_greynoise_403_error(
        self, mock_requests_get, domain_handler, test_data, capsys
    ):
        """
        Test exception behavior of Greynoise when not under any of the handled cases
        """
        handler = domain_handler(3)
        mock_resp = requests.models.Response()

        handler.resolutions = Resolutions.model_validate(
            json.loads(test_data("vt_resolutions_gist.json"))
        )

        mock_resp.status_code = 403
        mock_requests_get.return_value = mock_resp

        handler._fetch_greynoise(*["1.2.3.4", "1.2.3.5"])
        assert handler.warnings[0].startswith(
            "Could not fetch Greynoise: 403 Client Error:"
        )

        handler.print_warnings()
        capture = capsys.readouterr()
        assert capture.out.startswith(
            "WARN: Could not fetch Greynoise: 403 Client Error:"
        )

    @patch.object(requests.Session, "get")
    def test_greynoise_connection_error(
        self, mock_requests_get, domain_handler, test_data, capsys
    ):
        """
        Test exception behavior of Greynoise with non-HTTPError requests exception
        """
        handler = domain_handler(3)
        mock_requests_get.side_effect = ConnectionError("Foo bar message")

        handler.resolutions = Resolutions.model_validate(
            json.loads(test_data("vt_resolutions_gist.json"))
        )

        handler._fetch_greynoise(*["1.2.3.4", "1.2.3.5"])
        assert handler.warnings[0] == "Could not fetch Greynoise: Foo bar message"

        handler.print_warnings()
        capture = capsys.readouterr()
        assert capture.out.startswith(
            "WARN: Could not fetch Greynoise: Foo bar message"
        )

    @patch.object(requests.Session, "get")
    def test_greynoise_validation_error(self, mock_requests_get, domain_handler):
        handler = domain_handler()
        mock_resp = requests.models.Response()

        with patch.object(mock_resp, "json") as mock_resp_json:
            mock_resp.status_code = 200
            mock_resp_json.return_value = {"intentionally": "wrong data"}
            mock_requests_get.return_value = mock_resp

            with pytest.raises(HandlerException) as e:
                handler._fetch_greynoise("1.2.3.4")

            assert e.value.args[0].startswith(
                "Data model validation error: 5 validation errors for GreynoiseIp\n"
            )
            assert e.type is HandlerException

    @patch.object(requests.Session, "post")
    def test_urlhaus_http_error(self, mock_requests_post, domain_handler, capsys):
        """
        Test exception behavior of URLhaus with non-HTTPError requests exception
        """
        handler = domain_handler()
        mock_requests_post.side_effect = ConnectionError("Foo bar message")

        handler._fetch_urlhaus()
        assert handler.warnings[0] == "Could not fetch URLhaus: Foo bar message"

        handler.print_warnings()
        capture = capsys.readouterr()
        assert capture.out.startswith("WARN: Could not fetch URLhaus: Foo bar message")


class TestIpAddressHandler:
    def test_entity_refang(self, ip_handler):
        handler = ip_handler()
        assert handler.entity == "1.1.1.1"

    def test_fetch_data(self, ip_handler):
        handler = ip_handler()
        handler._fetch_vt_ip_address = MagicMock()
        handler._fetch_geoasn = MagicMock()
        handler._fetch_whois = MagicMock()
        handler._fetch_shodan = MagicMock()
        handler._fetch_greynoise = MagicMock()
        handler._fetch_urlhaus = MagicMock()
        handler._fetch_abuseipdb = MagicMock()

        fetch_data(MagicMock(), handler)
        handler._fetch_vt_ip_address.assert_called_once()
        handler._fetch_geoasn.assert_called_once()
        handler._fetch_whois.assert_called_once()
        handler._fetch_shodan.assert_called_once()
        handler._fetch_greynoise.assert_called_once()
        handler._fetch_urlhaus.assert_called_once()
        handler._fetch_abuseipdb.assert_called_once()

    @patch.object(requests.Session, "get")
    def test_vt_http_error(self, mock_requests_get, ip_handler):
        handler = ip_handler()
        mock_resp = requests.models.Response()

        mock_resp.status_code = 404
        mock_requests_get.return_value = mock_resp

        # Thorough test of first _fetch_* method
        with pytest.raises(HandlerException) as e:
            handler._fetch_vt_ip_address()

        assert (
            e.value.args[0]
            == "Error fetching data: 404 Client Error: None for url: None"
        )
        assert e.type is HandlerException

    @patch.object(requests.Session, "get")
    def test_vt_validation_error(self, mock_requests_get, ip_handler):
        handler = ip_handler()
        mock_resp = requests.models.Response()

        with patch.object(mock_resp, "json") as mock_resp_json:
            mock_resp.status_code = 200
            mock_resp_json.return_value = {"intentionally": "wrong data"}
            mock_requests_get.return_value = mock_resp

            # Thorough test of first _fetch_* method
            with pytest.raises(HandlerException) as e:
                handler._fetch_vt_ip_address()

            assert e.value.args[0].startswith(
                "Data model validation error: 1 validation error for IpAddress\ndata\n"
                "  Field required [type=missing, input_value={'intentionally': "
                "'wrong data'}, input_type=dict]\n"
            )
            assert e.type is HandlerException

    @patch.object(requests.Session, "get")
    def test_ipwhois_http_error(self, mock_requests_get, ip_handler, capsys):
        handler = ip_handler()

        mock_resp = requests.models.Response()

        mock_resp.status_code = 502
        mock_requests_get.return_value = mock_resp

        handler._fetch_geoasn("1.2.3.4")
        assert handler.warnings[0].startswith(
            "Could not fetch IPWhois: 502 Server Error:"
        )

        handler.print_warnings()
        capture = capsys.readouterr()
        assert capture.out.startswith(
            "WARN: Could not fetch IPWhois: 502 Server Error:"
        )

    @patch.object(requests.Session, "get")
    def test_whois_http_error(self, mock_requests_get, ip_handler, capsys):
        handler = ip_handler()
        mock_resp = requests.models.Response()

        mock_resp.status_code = 403
        mock_requests_get.return_value = mock_resp

        handler._fetch_whois()
        assert handler.warnings[0].startswith(
            "Could not fetch IP2Whois: 403 Client Error:"
        )

        handler.print_warnings()
        capture = capsys.readouterr()
        assert capture.out.startswith(
            "WARN: Could not fetch IP2Whois: 403 Client Error:"
        )

    @patch.object(requests.Session, "get")
    def test_ipwhois_validation_error(self, mock_requests_get, ip_handler):
        handler = ip_handler()
        mock_resp = requests.models.Response()

        with patch.object(mock_resp, "json") as mock_resp_json:
            mock_resp.status_code = 200
            mock_resp_json.return_value = {
                "success": True,
                "intentionally": "wrong data",
            }
            mock_requests_get.return_value = mock_resp

            with pytest.raises(HandlerException) as e:
                handler._fetch_geoasn("1.2.3.4")

            assert e.value.args[0].startswith(
                "Data model validation error: 1 validation error for IpWhois\n"
            )
            assert e.type is HandlerException

    @patch.object(requests.Session, "get")
    def test_shodan_http_error(self, mock_requests_get, ip_handler, capsys):
        """
        Test fail open behavior of Shodan on invalid API key
        """
        handler = ip_handler()
        mock_resp = requests.models.Response()

        mock_resp.status_code = 401
        mock_requests_get.return_value = mock_resp

        handler._fetch_shodan("1.2.3.4")
        assert handler.warnings[0].startswith(
            "Could not fetch Shodan: 401 Client Error:"
        )

        handler.print_warnings()
        capture = capsys.readouterr()
        assert capture.out.startswith("WARN: Could not fetch Shodan: 401 Client Error:")

    @patch.object(requests.Session, "get")
    def test_greynoise_http_error(self, mock_requests_get, ip_handler):
        """
        Test Greynoise HTTP error that results in a SystemExit
        """
        handler = ip_handler()
        mock_resp = requests.models.Response()

        mock_resp.status_code = 401
        mock_requests_get.return_value = mock_resp

        with pytest.raises(HandlerException) as e:
            handler._fetch_vt_ip_address()

        assert (
            e.value.args[0]
            == "Error fetching data: 401 Client Error: None for url: None"
        )
        assert e.type is HandlerException

    @patch.object(requests.Session, "get")
    def test_greynoise_429_error(self, mock_requests_get, ip_handler, capsys):
        """
        Test fail open behavior of Greynoise when rate limited
        """
        handler = ip_handler()
        mock_resp = requests.models.Response()

        mock_resp.status_code = 429
        mock_requests_get.return_value = mock_resp

        handler._fetch_greynoise("1.2.3.4")
        assert handler.warnings[0].startswith(
            "Could not fetch Greynoise: 429 Client Error:"
        )

        handler.print_warnings()
        capture = capsys.readouterr()
        assert capture.out.startswith(
            "WARN: Could not fetch Greynoise: 429 Client Error:"
        )

    @patch.object(requests.Session, "get")
    def test_greynoise_404_error(self, mock_requests_get, ip_handler):
        """
        Test fail open behavior of Greynoise when no IP found (404) and no warning
        message
        """
        handler = ip_handler()
        mock_resp = requests.models.Response()

        mock_resp.status_code = 404
        mock_requests_get.return_value = mock_resp

        handler._fetch_greynoise()
        assert len(handler.warnings) == 0
        assert handler.greynoise == GreynoiseIpMap.model_validate({})

    @patch.object(requests.Session, "get")
    def test_greynoise_403_error(self, mock_requests_get, ip_handler, capsys):
        """
        Test exception behavior of Greynoise when not under any of the handled cases
        """
        handler = ip_handler()
        mock_resp = requests.models.Response()

        mock_resp.status_code = 403
        mock_requests_get.return_value = mock_resp

        handler._fetch_greynoise("1.2.3.4")
        assert handler.warnings[0].startswith(
            "Could not fetch Greynoise: 403 Client Error:"
        )

        handler.print_warnings()
        capture = capsys.readouterr()
        assert capture.out.startswith(
            "WARN: Could not fetch Greynoise: 403 Client Error:"
        )

    @patch.object(requests.Session, "get")
    def test_greynoise_connection_error(self, mock_requests_get, ip_handler, capsys):
        """
        Test exception behavior of Greynoise with non-HTTPError requests exception
        """
        handler = ip_handler()
        mock_requests_get.side_effect = ConnectionError("Foo bar message")

        handler._fetch_greynoise("1.2.3.4")
        assert handler.warnings[0] == "Could not fetch Greynoise: Foo bar message"

        handler.print_warnings()
        capture = capsys.readouterr()
        assert capture.out.startswith(
            "WARN: Could not fetch Greynoise: Foo bar message"
        )

    @patch.object(requests.Session, "get")
    def test_greynoise_validation_error(self, mock_requests_get, ip_handler):
        handler = ip_handler()
        mock_resp = requests.models.Response()

        with patch.object(mock_resp, "json") as mock_resp_json:
            mock_resp.status_code = 200
            mock_resp_json.return_value = {"intentionally": "wrong data"}
            mock_requests_get.return_value = mock_resp

            with pytest.raises(HandlerException) as e:
                handler._fetch_greynoise("1.2.3.4")

            assert e.value.args[0].startswith(
                "Data model validation error: 5 validation errors for GreynoiseIp\n"
            )
            assert e.type is HandlerException

    @patch.object(requests.Session, "post")
    def test_urlhaus_http_error(self, mock_requests_post, ip_handler, capsys):
        """
        Test exception behavior of URLhaus with non-HTTPError requests exception
        """
        handler = ip_handler()
        mock_requests_post.side_effect = ConnectionError("Foo bar message")

        handler._fetch_urlhaus()
        assert handler.warnings[0] == "Could not fetch URLhaus: Foo bar message"

        handler.print_warnings()
        capture = capsys.readouterr()
        assert capture.out.startswith("WARN: Could not fetch URLhaus: Foo bar message")

    @patch.object(requests.Session, "get")
    def test_abuseipdb_429_error(self, mock_requests_get, ip_handler, capsys):
        """
        Test fail open behavior of AbuseIPDB when rate limited
        """
        handler = ip_handler()
        mock_resp = requests.models.Response()

        mock_resp.status_code = 429
        mock_requests_get.return_value = mock_resp

        handler._fetch_abuseipdb("1.2.3.4")
        assert handler.warnings[0].startswith(
            "Could not fetch AbuseIPDB: 429 Client Error:"
        )

        handler.print_warnings()
        capture = capsys.readouterr()
        assert capture.out.startswith(
            "WARN: Could not fetch AbuseIPDB: 429 Client Error:"
        )



import pytest
from freezegun import freeze_time
from rich.text import Text

from wtfis.utils import Timestamp, error_and_exit, is_ip, older_than, refang, smart_join


class TestUtils:
    def test_Timestamp_1(self):
        assert Timestamp(1660428690).timestamp == "2022-08-13T22:11:30Z"

    def test_Timestamp_2(self):
        assert (
            Timestamp("2017-04-26T15:43:49.000-07:00").timestamp
            == "2017-04-26T22:43:49Z"
        )

    def test_Timestamp_3(self):
        assert Timestamp(None).timestamp is None
        assert Timestamp("2017-04-26T22:43:49Z").timestamp == "2017-04-26T22:43:49Z"
        assert Timestamp("1660428690").timestamp == "2022-08-13T22:11:30Z"

    def test_Timestamp_4(self):
        assert Timestamp("2017-04-26T22:43:49.00Z").timestamp == "2017-04-26T22:43:49Z"
        assert Timestamp("2017-04-26T22:43:49+0000").timestamp == "2017-04-26T22:43:49Z"
        assert Timestamp("2017-04-26 22:43:49").timestamp == "2017-04-26 22:43:49"
        assert Timestamp("a long time ago").timestamp == "a long time ago"

    @freeze_time("2022-08-12")
    def test_older_than(self):
        assert older_than(1659146024, 30) is False
        assert older_than(1657418024, 30) is True

    def test_smart_join_1(self):
        output = smart_join(*["foo", "bar", "baz"])
        assert type(output) is Text
        assert str(output) == "foo, bar, baz"
        assert output == (
            Text()
            .append("foo")
            .append(", ", style="default")
            .append("bar")
            .append(", ", style="default")
            .append("baz")
        )

    def test_smart_join_2(self):
        output = smart_join(*["foo", "bar", "baz"], style="blue")
        assert type(output) is Text
        assert str(output) == "foo, bar, baz"
        assert output == (
            Text()
            .append("foo", style="blue")
            .append(", ", style="default")
            .append("bar", style="blue")
            .append(", ", style="default")
            .append("baz", style="blue")
        )

    def test_smart_join_3(self):
        output_1 = smart_join(*[])
        assert type(output_1) is str
        assert output_1 == ""

        output_2 = smart_join(None)
        assert type(output_2) is str
        assert output_2 == ""

    def test_error_and_exit(self):
        with pytest.raises(SystemExit) as e:
            error_and_exit("foo", 2)
        assert e.value.code == 2

    def test_refang(self):
        assert refang("foo[.]bar[.]com") == "foo.bar.com"

    def test_is_ip(self):
        assert is_ip("1.1.1.1") is True
        assert is_ip("foo.example.com") is False



import re
import sys
from datetime import datetime, timedelta, timezone
from ipaddress import ip_address
from typing import Optional, Union

from rich.console import RenderableType
from rich.text import Text

from wtfis.ui.theme import Theme


class Timestamp:
    std_utc = "%Y-%m-%dT%H:%M:%SZ"
    ts_regex = re.compile(r"(?P<date>[0-9-]+)T(?P<time>[0-9:]+)Z")
    theme = Theme()

    def __init__(self, ts: Union[str, int, None]):
        self.timestamp = self._standardize(ts)

    def _standardize(self, ts: Union[str, int, None]) -> Optional[str]:
        """Convert any time to a standard format"""
        # Do nothing if ts is None
        if ts is None:
            return None

        # Return as is if format is already what we want
        try:
            datetime.strptime(str(ts), self.std_utc)
            return str(ts)
        except ValueError:
            pass

        # Attempt to detect epoch time in string format
        try:
            ts = int(ts)
        except ValueError:
            pass

        if isinstance(ts, int):
            return datetime.fromtimestamp(ts, tz=timezone.utc).strftime(self.std_utc)
        elif isinstance(ts, str):
            # Try to convert other possible date formats
            for fmt in [
                "%Y-%m-%dT%H:%M:%S.00Z",
                "%Y-%m-%dT%H:%M:%S%z",
            ]:
                try:
                    return datetime.strptime(ts, fmt).strftime(self.std_utc)
                except ValueError:
                    pass

            # No explicit timezone - return date as is (le sigh)
            for fmt in [
                "%Y-%m-%d %H:%M:%S",
            ]:
                try:
                    datetime.strptime(ts, fmt)
                    return ts
                except ValueError:
                    pass

            # Default
            try:
                return (
                    datetime.fromisoformat(ts)
                    .astimezone(timezone.utc)
                    .strftime(self.std_utc)
                )
            except ValueError:  # Cannot convert; fail open
                return ts

    @property
    def render(self) -> Optional[RenderableType]:
        if self.timestamp is None:  # Return None if no ts to begin with
            return None

        match = self.ts_regex.match(self.timestamp) if self.timestamp else None
        if not match:  # Return ts as is if it does not match our regex
            return self.timestamp

        text = Text(match.group("date"), style=self.theme.timestamp_date)
        text.append("T", style=self.theme.timestamp_t)
        text.append(match.group("time"), style=self.theme.timestamp_time)
        text.append("Z", style=self.theme.timestamp_z)
        return text


def older_than(ts: int, days: int) -> bool:
    """Tells if a timestamp is older than X days. "ts" must be epoch format"""
    return datetime.fromtimestamp(ts) < datetime.now() - timedelta(days=days)


def smart_join(
    *items: Optional[Union[Text, str]], style: Optional[str] = None
) -> Union[Text, str]:
    text = Text()
    for idx, item in enumerate(items):
        if item:
            if isinstance(item, Text):
                text.append(item)
            else:
                text.append(item, style=style)
            if idx < len(items) - 1:
                text.append(", ", style="default")
    return text if str(text) != "" else ""


def error_and_exit(message: str, status: int = 1):
    print(message, file=sys.stderr)
    raise SystemExit(status)


def refang(text: str) -> str:
    """Strip []s out of text"""
    return text.replace("[", "").replace("]", "")


def is_ip(text: str) -> bool:
    """Detect whether text is IPv4 or not"""
    try:
        return ip_address(refang(text)).is_global
    except ValueError:
        return False



from dataclasses import dataclass
from typing import List, Optional, Tuple, Union

from rich.columns import Columns
from rich.console import Group, RenderableType
from rich.padding import Padding
from rich.panel import Panel
from rich.text import Text

from wtfis.models.virustotal import Resolutions
from wtfis.ui.base import BaseView
from wtfis.utils import Timestamp, smart_join


@dataclass
class DomainView(BaseView):
    """
    Handler for FQDN and domain lookup output
    """

    resolutions: Optional[Resolutions]
    max_resolutions: int = 3

    def domain_panel(self) -> Panel:
        content = [self._gen_vt_section()]  # VT section
        for section in (self._gen_urlhaus_section(),):  # URLhaus section
            if section is not None:
                content.append("")
                content.append(section)

        return self._gen_panel(self._gen_group(content), self.entity.data.id_)

    def resolutions_panel(self) -> Optional[Panel]:
        # Skip if no resolutions data
        if not self.resolutions:
            return None

        content: List[RenderableType] = [self._gen_heading_text("Resolutions")]

        for idx, ip in enumerate(self.resolutions.data):
            if idx == self.max_resolutions:
                break
            attributes = ip.attributes

            # Analysis
            analysis = self._gen_vt_analysis_stats(
                attributes.ip_address_last_analysis_stats
            )
            analysis_field = self._gen_linked_field_name(
                "Analysis",
                hyperlink=f"{self.vt_gui_baseurl_ip}/{attributes.ip_address}",
            )

            # Content
            heading = self._gen_heading_text(
                attributes.ip_address,
                # hyperlink=f"{self.vt_gui_baseurl_ip}/{attributes.ip_address}",
                type="h2",
            )
            data: List[Tuple[Union[str, Text], Union[RenderableType, None]]] = [
                (analysis_field, analysis),
                ("Resolved:", Timestamp(attributes.date).render),
            ]

            # IP geolocation and ASN
            geoasn = self._get_geoasn_enrichment(attributes.ip_address)
            if geoasn:
                asn_field: Union[Text, str] = (
                    self._gen_linked_field_name("ASN", hyperlink=geoasn.link)
                    if geoasn.link
                    else "ASN:"
                )
                asn = self._gen_asn_text(geoasn.asn, geoasn.org)
                data += [
                    (asn_field, asn),
                    ("ISP:", geoasn.isp),
                    (
                        "Location:",
                        smart_join(geoasn.city, geoasn.region, geoasn.country),
                    ),
                    ("Hostname:", geoasn.hostname),
                    ("Proxy:", geoasn.is_proxy),
                    ("Anycast:", geoasn.is_anycast),
                ]

            # Shodan
            shodan = self._get_shodan_enrichment(attributes.ip_address)
            if shodan:
                tags = (
                    smart_join(*shodan.tags, style=self.theme.tags)
                    if shodan.tags
                    else None
                )
                services_field = self._gen_linked_field_name(
                    "Services",
                    hyperlink=f"{self.shodan_gui_baseurl}/{attributes.ip_address}",
                )
                data += [
                    ("OS:", shodan.os),
                    (services_field, self._gen_shodan_services(shodan)),
                    ("Tags:", tags),
                    # Timestamps are UTC (source: Google)
                    # ("Last Scan:", Timestamp(f"{shodan.last_update}+00:00").render),
                ]

            # Greynoise
            greynoise = self._get_greynoise_enrichment(attributes.ip_address)
            if greynoise:
                data += [self._gen_greynoise_tuple(greynoise)]

            # AbuseIPDB
            abuseipdb = self._get_abuseipdb_enrichment(attributes.ip_address)
            if abuseipdb:
                data += [self._gen_abuseipdb_tuple(abuseipdb)]

            # Include a disclaimer if last seen is older than 1 year
            # Note: Disabled for now because I originally understood that the
            # resolution date was the last time the domain was resolved, but it may
            # actually be he first time the IP itself was seen with the domain.
            # if enrich and older_than(attributes.date, 365):
            #     body = Group(
            #         self._gen_table(*data),
            #         Text(
            #             "**Enrichment data may be inaccurate",
            #              style=self.theme.disclaimer
            #         ),
            #     )  # type: Union[Group, Table, str]
            # else:
            #     body = self._gen_table(*data)

            body = self._gen_table(*data)
            content.append(self._gen_section(body, heading))

            # Add extra line break if not last item in list
            if idx < self.max_resolutions - 1 and idx < len(self.resolutions.data) - 1:
                content.append("")

        # Info about how many more IPs were not shown
        if self.max_resolutions < self.resolutions.meta.count:
            content.append("")
            content.append(
                Text(justify="center", end="\n").append(
                    f"+{self.resolutions.meta.count - self.max_resolutions} more",
                    style=(
                        f"{self.theme.footer} "
                        f"link {self.vt_gui_baseurl_domain}/{self.entity.data.id_}"
                        "/relations"
                    ),
                )
            )

        # Render results, if existent
        if len(content) > 1:
            return self._gen_panel(self._gen_group(content))

        # No result
        return None

    def print(self, one_column: bool = False) -> None:
        renderables = [
            i
            for i in (
                self.domain_panel(),
                self.resolutions_panel(),
                self.whois_panel(),
            )
            if i is not None
        ]

        if one_column:
            self.console.print(Group(*([""] + renderables)))  # type: ignore
        else:
            self.console.print(Padding(Columns(renderables), (1, 0)))


class IpAddressView(BaseView):
    """
    Handler for IP Address lookup output
    """

    def ip_panel(self) -> Panel:
        content = [self._gen_vt_section()]  # VT section
        for section in (
            self._gen_geoasn_section(),  # IP location and ASN section
            self._gen_shodan_section(),  # Shodan section
            self._gen_urlhaus_section(),  # URLhaus section
            self._gen_ip_other_section(),  # Other section
        ):
            if section is not None:
                content.append("")
                content.append(section)

        return self._gen_panel(self._gen_group(content), self.entity.data.id_)

    def print(self, one_column: bool = False) -> None:
        renderables = [
            i
            for i in (
                self.ip_panel(),
                self.whois_panel(),
            )
            if i is not None
        ]

        if one_column:
            self.console.print(Group(*([""] + renderables)))  # type: ignore
        else:
            self.console.print(Padding(Columns(renderables), (1, 0)))



from typing import Optional

from rich.console import Console
from rich.progress import Progress, TaskID

from wtfis.config import Config
from wtfis.exceptions import HandlerException, WtfisException
from wtfis.handlers.base import BaseHandler
from wtfis.handlers.domain import DomainHandler
from wtfis.handlers.ip import IpAddressHandler
from wtfis.models.virustotal import Domain, IpAddress
from wtfis.ui.base import BaseView
from wtfis.ui.progress import get_progress
from wtfis.ui.view import DomainView, IpAddressView
from wtfis.utils import error_and_exit, is_ip


def generate_entity_handler(
    config: Config,
    console: Console,
) -> BaseHandler:
    # Domain / FQDN handler
    if not is_ip(config.entity):
        entity: BaseHandler = DomainHandler(
            entity=config.entity,
            console=console,
            vt_client=config.vt_client,
            ip_geoasn_client=config.ip_geoasn_client,
            whois_client=config.whois_client,
            shodan_client=config.shodan_client,
            greynoise_client=config.greynoise_client,
            abuseipdb_client=config.abuseipdb_client,
            urlhaus_client=config.urlhaus_client,
            max_resolutions=config.max_resolutions,
        )
    # IP address handler
    else:
        entity = IpAddressHandler(
            entity=config.entity,
            console=console,
            vt_client=config.vt_client,
            ip_geoasn_client=config.ip_geoasn_client,
            whois_client=config.whois_client,
            shodan_client=config.shodan_client,
            greynoise_client=config.greynoise_client,
            abuseipdb_client=config.abuseipdb_client,
            urlhaus_client=config.urlhaus_client,
        )

    return entity


def generate_view(
    config: Config,
    console: Console,
    entity: BaseHandler,
) -> BaseView:
    # Output display
    if isinstance(entity, DomainHandler) and isinstance(entity.vt_info, Domain):
        view: BaseView = DomainView(
            console,
            entity.vt_info,
            entity.geoasn,
            entity.whois,
            entity.shodan,
            entity.greynoise,
            entity.abuseipdb,
            entity.urlhaus,
            entity.resolutions,
            max_resolutions=config.max_resolutions,
        )
    elif isinstance(entity, IpAddressHandler) and isinstance(entity.vt_info, IpAddress):
        view = IpAddressView(
            console,
            entity.vt_info,
            entity.geoasn,
            entity.whois,
            entity.shodan,
            entity.greynoise,
            entity.abuseipdb,
            entity.urlhaus,
        )
    else:
        raise WtfisException("Unsupported entity!")

    return view


def fetch_data(
    progress: Progress,
    entity: BaseHandler,
):

    def _finish_task():
        if task is not None:
            progress.update(task, completed=100)

    task: Optional[TaskID] = None
    with progress:
        try:
            for x in entity.fetch_data():
                if isinstance(x, tuple):  # (str, int)
                    _finish_task()
                    descr, adv = x
                    task = progress.add_task(descr)
                    progress.update(task, advance=adv)
                elif isinstance(x, int) and task is not None:
                    progress.update(task, advance=x)
            _finish_task()
        except HandlerException as e:
            progress.stop()
            error_and_exit(str(e))


def main():
    # Load config
    config = Config()

    # Instantiate the console
    console = Console(no_color=True) if config.no_color else Console()

    # Progress animation controller
    progress = get_progress(console)

    # Entity handler
    entity = generate_entity_handler(config, console)

    # Fetch data
    fetch_data(progress, entity)

    # Print fetch warnings, if any
    entity.print_warnings()

    # Output display
    view = generate_view(config, console, entity)

    # Finally, print output
    view.print(one_column=config.one_column)



"""
ip2location datamodels
API doc: https://www.ip2location.io/ip2location-documentation
"""

from __future__ import annotations

from typing import Dict, Optional

from pydantic import ConfigDict, Field

from wtfis.models.base import IpGeoAsnBase, IpGeoAsnMapBase


class Ip2Location(IpGeoAsnBase):
    # Metadata
    source: str = "IP2Location"

    # Config
    model_config = ConfigDict(populate_by_name=True)

    # Results
    city: Optional[str] = Field(None, alias="city_name")
    region: Optional[str] = Field(None, alias="region_name")
    country: Optional[str] = Field(None, alias="country_name")
    org: Optional[str] = Field(None, alias="as")


class Ip2LocationMap(IpGeoAsnMapBase):  # type: ignore[override]
    root: Dict[str, Ip2Location]

    @classmethod
    def empty(cls) -> Ip2LocationMap:
        return cls.model_validate({})



"""
Type aliases
"""

from typing import Union

from wtfis.models.abuseipdb import AbuseIpDbMap
from wtfis.models.greynoise import GreynoiseIpMap
from wtfis.models.ip2location import Ip2Location, Ip2LocationMap
from wtfis.models.ipinfo import IpInfo, IpInfoMap
from wtfis.models.ipwhois import IpWhois, IpWhoisMap
from wtfis.models.shodan import ShodanIpMap
from wtfis.models.urlhaus import UrlHausMap

# IP enrichment map types
IpEnrichmentType = Union[
    AbuseIpDbMap,
    GreynoiseIpMap,
    Ip2LocationMap,
    IpInfoMap,
    IpWhoisMap,
    ShodanIpMap,
    UrlHausMap,
]

# Domain/FQDN enrichment map types
DomainEnrichmentType = Union[UrlHausMap,]

# IP geolocation and ASN types
IpGeoAsnType = Union[
    Ip2Location,
    IpInfo,
    IpWhois,
]

IpGeoAsnMapType = Union[
    Ip2LocationMap,
    IpInfoMap,
    IpWhoisMap,
]



import json
from unittest.mock import MagicMock, patch

import pytest

from wtfis.clients.abuseipdb import AbuseIpDbClient
from wtfis.clients.base import requests
from wtfis.clients.greynoise import GreynoiseClient
from wtfis.clients.ip2location import Ip2LocationClient
from wtfis.clients.ip2whois import Ip2WhoisClient
from wtfis.clients.ipinfo import IpInfoClient
from wtfis.clients.ipwhois import IpWhoisClient
from wtfis.clients.shodan import ShodanClient
from wtfis.clients.urlhaus import UrlHausClient
from wtfis.clients.virustotal import VTClient
from wtfis.models.ipwhois import IpWhoisMap
from wtfis.models.shodan import ShodanIpMap


@pytest.fixture()
def abuseipdb_client():
    return AbuseIpDbClient("dummykey")


@pytest.fixture()
def greynoise_client():
    return GreynoiseClient("dummykey")


@pytest.fixture()
def ip2location_client():
    return Ip2LocationClient("dummykey")


@pytest.fixture()
def ip2whois_client():
    return Ip2WhoisClient("dummykey")


@pytest.fixture()
def ipinfo_client():
    return IpInfoClient()


@pytest.fixture()
def ipwhois_client():
    return IpWhoisClient()


@pytest.fixture()
def shodan_client():
    return ShodanClient("dummykey")


@pytest.fixture()
def urlhaus_client():
    return UrlHausClient("dummykey")


@pytest.fixture()
def virustotal_client():
    return VTClient("dummykey")


class TestAbuseIpDbClient:
    def test_init(self, abuseipdb_client):
        assert abuseipdb_client.name == "AbuseIPDB"
        assert abuseipdb_client.api_key == "dummykey"

    @patch.object(requests.Session, "get")
    def test_enrich_ips(self, mock_requests_get, test_data, abuseipdb_client):
        mock_resp = MagicMock()
        mock_resp.status_code = 200
        mock_resp.json.return_value = json.loads(
            test_data("abuseipdb_1.1.1.1_raw.json")
        )
        mock_requests_get.return_value = mock_resp

        abuseipdb = abuseipdb_client.enrich_ips("thisdoesntmatter").root["1.1.1.1"]

        assert abuseipdb.ip_address == "1.1.1.1"
        assert abuseipdb.abuse_confidence_score == 100
        assert abuseipdb.usage_type == "Data Center/Web Hosting/Transit"
        assert abuseipdb.total_reports == 567
        assert abuseipdb.num_distinct_users == 123


class TestGreynoiseClient:
    def test_init(self, greynoise_client):
        assert greynoise_client.name == "Greynoise"
        assert greynoise_client.api_key == "dummykey"


class TestIp2LocationClient:
    def test_init(self, ip2location_client):
        assert ip2location_client.name == "IP2Location"
        assert ip2location_client.api_key == "dummykey"

    @patch.object(requests.Session, "get")
    def test_enrich_ips(self, mock_requests_get, test_data, ip2location_client):
        mock_resp = MagicMock()
        mock_resp.status_code = 200
        mock_resp.json.return_value = json.loads(
            test_data("ip2location_1.1.1.1.json")
        ).get("1.1.1.1")
        mock_requests_get.return_value = mock_resp

        ip2location = ip2location_client.enrich_ips("1.1.1.1").root["1.1.1.1"]

        assert ip2location.ip == "1.1.1.1"
        assert ip2location.city == "Brisbane"
        assert ip2location.region == "Queensland"
        assert ip2location.country == "Australia"
        assert ip2location.org == "CloudFlare Inc."
        assert ip2location.is_proxy == "False"


class TestIp2WhoisClient:
    def test_init(self, ip2whois_client):
        assert ip2whois_client.api_key == "dummykey"
        assert ip2whois_client.name == "IP2Whois"

    @patch.object(requests.Session, "get")
    def test_get_whois(self, mock_requests_get, test_data, ip2whois_client):
        mock_resp = MagicMock()
        mock_resp.status_code = 200
        mock_resp.json.return_value = json.loads(
            test_data("ip2whois_whois_hotmail.json")
        )
        mock_requests_get.return_value = mock_resp

        whois = ip2whois_client.get_whois("thisdoesntmatter")

        assert whois.domain == "hotmail.com"

    @patch.object(requests.Session, "get")
    def test_get_whois_http_error(self, mock_requests_get, ip2whois_client):
        mock_resp = requests.models.Response()

        with patch.object(mock_resp, "json") as mock_resp_json:
            # 404 - no exception
            mock_resp.status_code = 404
            mock_resp_json.return_value = {"error": {"error_code": 10006}}
            mock_requests_get.return_value = mock_resp
            assert ip2whois_client.get_whois("thisdoesntmatter").source == "ip2whois"

            # 400 - no exception
            mock_resp.status_code = 400
            mock_resp_json.return_value = {"error": {"error_code": 10007}}
            mock_requests_get.return_value = mock_resp
            assert ip2whois_client.get_whois("thisdoesntmatter").source == "ip2whois"

            # 400 - with exception
            mock_resp.status_code = 400
            mock_resp_json.return_value = {"error": {"error_code": 10008}}
            mock_requests_get.return_value = mock_resp
            with pytest.raises(requests.exceptions.HTTPError) as err:
                ip2whois_client.get_whois("thisdoesntmatter")
            assert err.value.response.json()["error"]["error_code"] == 10008


class TestIpInfoClient:
    def test_init(self, ipinfo_client):
        assert ipinfo_client.name == "IPinfo"

    @patch.object(requests.Session, "get")
    def test_enrich_ips(self, mock_requests_get, test_data, ipinfo_client):
        mock_resp = MagicMock()
        mock_resp.status_code = 200
        mock_resp.json.return_value = json.loads(test_data("ipinfo_1.1.1.1.json")).get(
            "1.1.1.1"
        )
        mock_requests_get.return_value = mock_resp

        ipinfo = ipinfo_client.enrich_ips("1.1.1.1").root["1.1.1.1"]

        assert ipinfo.ip == "1.1.1.1"
        assert ipinfo.city == "Brisbane"
        assert ipinfo.region == "Queensland"
        assert ipinfo.country == "AU"
        assert ipinfo.org == "Cloudflare, Inc."
        assert ipinfo.is_anycast == "True"
        assert ipinfo.link == "https://ipinfo.io/1.1.1.1"


class TestIpWhoisClient:
    def test_init(self, ipwhois_client):
        assert ipwhois_client.name == "IPWhois"

    @patch.object(requests.Session, "get")
    def test_get_ipwhois(self, mock_requests_get, ipwhois_client):
        mock_resp = MagicMock()
        mock_resp.status_code = 200
        mock_resp.json.return_value = json.loads("{}")
        mock_requests_get.return_value = mock_resp

        whois = ipwhois_client.enrich_ips("thisdoesntmatter")

        assert whois == IpWhoisMap.model_validate({})


class TestShodanClient:
    def test_init(self, shodan_client):
        assert shodan_client.api_key == "dummykey"
        assert shodan_client.name == "Shodan"

    @patch.object(requests.Session, "get")
    def test_enrich_ips_ok(self, mock_requests_get, test_data, shodan_client):
        mock_resp = MagicMock()
        mock_resp.status_code = 200
        mock_resp.json.return_value = json.loads(test_data("shodan_1.1.1.1.json"))[
            "1.1.1.1"
        ]
        mock_requests_get.return_value = mock_resp

        shodan = shodan_client.enrich_ips("thisdoesntmatter").root["1.1.1.1"]

        assert len(shodan.data) == 13
        assert shodan.ports == [
            161,
            2082,
            2083,
            69,
            2086,
            2087,
            80,
            8880,
            8080,
            53,
            8443,
            443,
        ]
        assert shodan.tags == []

    @patch.object(requests.Session, "get")
    def test_enrich_ips_404(self, mock_requests_get, shodan_client):
        mock_resp = requests.models.Response()
        mock_resp.status_code = 404
        mock_requests_get.return_value = mock_resp
        assert shodan_client.enrich_ips("thisdoesntmatter") == ShodanIpMap.empty()

    @patch.object(requests.Session, "get")
    def test_enrich_ips_401(self, mock_requests_get, shodan_client):
        mock_resp = requests.models.Response()
        mock_resp.status_code = 401
        mock_requests_get.return_value = mock_resp

        with pytest.raises(requests.exceptions.HTTPError) as err:
            shodan_client.enrich_ips("thisdoesntmatter")

        assert err.value.response.status_code == 401


class TestUrlhausClient:
    def test_init(self, urlhaus_client):
        assert urlhaus_client.name == "URLhaus"

    @patch.object(requests.Session, "post")
    def test_enrich_ips(self, mock_requests_post, test_data, urlhaus_client):
        mock_resp = MagicMock()
        mock_resp.status_code = 200
        mock_resp.json.return_value = json.loads(test_data("urlhaus_1.1.1.1.json"))[
            "1.1.1.1"
        ]
        mock_requests_post.return_value = mock_resp

        urlhaus = urlhaus_client.enrich_ips("thisdoesntmatter").root["1.1.1.1"]

        assert urlhaus.host == "1.1.1.1"
        assert urlhaus.url_count == 10
        assert urlhaus.urlhaus_reference == "https://urlhaus.abuse.ch/host/1.1.1.1/"


class TestVirustotalClient:
    def test_init(self, virustotal_client):
        assert virustotal_client.s.headers["x-apikey"] == "dummykey"
        assert virustotal_client.s.headers["Accept"] == "application/json"
        assert virustotal_client.name == "Virustotal"

    @patch.object(requests.Session, "get")
    def test_get_whois_domain(self, mock_requests_get, test_data, virustotal_client):
        mock_resp = MagicMock()
        mock_resp.status_code = 200
        mock_resp.json.return_value = json.loads(test_data("vt_whois_gist.json"))
        mock_requests_get.return_value = mock_resp

        whois = virustotal_client.get_whois("gist.github.com")

        assert whois.domain == "github.com"
        assert whois.registrar == "MarkMonitor Inc."

    @patch.object(requests.Session, "get")
    def test_get_whois_ip(self, mock_requests_get, test_data, virustotal_client):
        mock_resp = MagicMock()
        mock_resp.status_code = 200
        mock_resp.json.return_value = json.loads(test_data("vt_whois_1.1.1.1.json"))
        mock_requests_get.return_value = mock_resp

        whois = virustotal_client.get_whois("1.1.1.1")

        assert whois.domain == "one.one"
        assert whois.registrar == "One.com A/S - ONE"



from typing import Optional

from wtfis.clients.base import BaseIpEnricherClient, BaseRequestsClient
from wtfis.models.ip2location import Ip2Location, Ip2LocationMap


class Ip2LocationClient(BaseRequestsClient, BaseIpEnricherClient):
    """
    IP2Location client
    """

    baseurl = "https://api.ip2location.io"

    def __init__(self, api_key: str) -> None:
        super().__init__()
        self.api_key = api_key

    @property
    def name(self) -> str:
        return "IP2Location"

    def _get_ip2location(self, ip: str) -> Optional[Ip2Location]:
        params = {
            "key": self.api_key,
            "ip": ip,
            "format": "json",
        }
        return Ip2Location.model_validate(self._get("/", params=params))

    def enrich_ips(self, *ips: str) -> Ip2LocationMap:
        map_ = {}
        for ip in ips:
            ip2location = self._get_ip2location(ip)
            if ip2location:
                map_[ip2location.ip] = ip2location
        return Ip2LocationMap.model_validate(map_)



from typing import Optional

from wtfis.clients.base import BaseIpEnricherClient, BaseRequestsClient
from wtfis.models.abuseipdb import AbuseIpDb, AbuseIpDbMap


class AbuseIpDbClient(BaseRequestsClient, BaseIpEnricherClient):
    """
    AbuseIPDB client
    """

    baseurl = "https://api.abuseipdb.com/api/v2"

    def __init__(self, api_key: str) -> None:
        super().__init__()
        self.api_key = api_key

    @property
    def name(self) -> str:
        return "AbuseIPDB"

    def _get_ip(self, ip: str) -> Optional[AbuseIpDb]:
        params = {"ipAddress": ip, "maxAgeInDays": "90"}
        headers = {"key": self.api_key, "Accept": "application/json"}

        response = self._get(request="/check", headers=headers, params=params)

        return AbuseIpDb.model_validate(response["data"])

    def enrich_ips(self, *ips: str) -> AbuseIpDbMap:
        abuseipdb_map = {}
        for ip in ips:
            ip_data = self._get_ip(ip)
            if ip_data:
                abuseipdb_map[ip_data.ip_address] = ip_data
        return AbuseIpDbMap.model_validate(abuseipdb_map)



from wtfis.clients.base import (
    BaseDomainEnricherClient,
    BaseIpEnricherClient,
    BaseRequestsClient,
)
from wtfis.models.urlhaus import UrlHaus, UrlHausMap


class UrlHausClient(BaseRequestsClient, BaseDomainEnricherClient, BaseIpEnricherClient):
    """
    URLhaus client
    """

    baseurl = "https://urlhaus-api.abuse.ch/v1"

    def __init__(self, api_key: str) -> None:
        super().__init__()
        self.api_key = api_key

    @property
    def name(self) -> str:
        return "URLhaus"

    def _get_host(self, host: str) -> UrlHaus:
        data = {"host": host}
        headers = {"Auth-Key": self.api_key}
        return UrlHaus.model_validate(self._post("/host", data=data, headers=headers))

    def _enrich(self, *entities: str) -> UrlHausMap:
        """Method is the same whether input is a domain or IP"""
        urlhaus_map = {}
        for entity in entities:
            data = self._get_host(entity)
            if data.host:
                urlhaus_map[data.host] = data
        return UrlHausMap.model_validate(urlhaus_map)

    def enrich_domains(self, *domains: str) -> UrlHausMap:
        return self._enrich(*domains)

    def enrich_ips(self, *ips: str) -> UrlHausMap:
        return self._enrich(*ips)



from typing import Optional

from requests.exceptions import HTTPError

from wtfis.clients.base import BaseIpEnricherClient, BaseRequestsClient
from wtfis.models.shodan import ShodanIp, ShodanIpMap


class ShodanClient(BaseRequestsClient, BaseIpEnricherClient):
    """
    Shodan client
    """

    baseurl = "https://api.shodan.io/shodan"

    def __init__(self, api_key: str) -> None:
        super().__init__()
        self.api_key = api_key

    @property
    def name(self) -> str:
        return "Shodan"

    def _get_ip(self, ip: str) -> Optional[ShodanIp]:
        params = {"key": self.api_key}
        try:
            return ShodanIp.model_validate(self._get(f"/host/{ip}", params=params))
        except HTTPError as e:
            # 404 means the IP is invalid or not in Shodan; we don't want to raise an error
            if e.response.status_code == 404:
                return None
            raise e

    def enrich_ips(self, *ips: str) -> ShodanIpMap:
        shodan_map = {}
        for ip in ips:
            ip_data = self._get_ip(ip)
            if ip_data:
                shodan_map[ip_data.ip_str] = ip_data
        return ShodanIpMap.model_validate(shodan_map)



from typing import Any, Dict, List, Optional, Set

from pydantic import BaseModel, Field, RootModel, field_validator, model_validator

from wtfis.models.base import LaxStr, WhoisBase


class BaseData(BaseModel):
    attributes: Any = None
    id_: str = Field(alias="id")
    type_: str = Field(alias="type")


class Meta(BaseModel):
    count: int


class AnalysisResult(BaseModel):
    category: str
    engine_name: str
    method: str
    result: str


class LastAnalysisResults(RootModel):
    root: Dict[str, AnalysisResult]


class LastAnalysisStats(BaseModel):
    harmless: int
    malicious: int
    suspicious: int
    timeout: int
    undetected: int


class Popularity(BaseModel):
    rank: int
    timestamp: int


class PopularityRanks(RootModel):
    root: Dict[str, Popularity]


class BaseAttributes(BaseModel):
    jarm: Optional[str] = None
    last_analysis_results: LastAnalysisResults
    last_analysis_stats: LastAnalysisStats
    last_https_certificate_date: Optional[int] = None
    last_modification_date: Optional[int] = None
    reputation: int
    tags: List[str]


class DomainAttributes(BaseAttributes):
    categories: List[str]
    creation_date: Optional[int] = None
    last_dns_records_date: Optional[int] = None
    last_update_date: Optional[int] = None
    popularity_ranks: PopularityRanks
    registrar: Optional[str] = None

    @field_validator("categories", mode="before")
    @classmethod
    def transform_categories(cls, v):
        cats: Set[str] = set()
        for category in v.values():
            for delimiter in [", ", ",", "/", " / "]:
                if delimiter in category:
                    cats = cats | set(category.lower().split(delimiter))
                    break
            else:
                cats.add(category.lower())
        return sorted(cats)


class DomainData(BaseData):
    attributes: DomainAttributes


class Domain(BaseModel):
    data: DomainData


class IpAttributes(BaseAttributes):
    asn: Optional[LaxStr] = None
    continent: Optional[str] = None
    country: Optional[str] = None
    network: Optional[str] = None


class IpData(BaseData):
    attributes: IpAttributes


class IpAddress(BaseModel):
    data: IpData


class ResolutionAttributes(BaseModel):
    date: int
    host_name: str
    resolver: str
    ip_address_last_analysis_stats: LastAnalysisStats
    ip_address: str
    host_name_last_analysis_stats: LastAnalysisStats


class ResolutionData(BaseData):
    attributes: ResolutionAttributes


class Resolutions(BaseModel):
    meta: Meta
    data: List[ResolutionData]

    def ip_list(self, limit: int) -> List[str]:
        """
        IP list from the first n resolutions
        where n is defined by the limit
        """
        ips = []
        for idx, resolution in enumerate(self.data):
            if idx < limit:
                ips.append(resolution.attributes.ip_address)
        return ips


class Whois(WhoisBase):
    source: str = "virustotal"
    domain: str = Field("", alias="Domain Name")
    registrar: Optional[str] = None
    organization: Optional[str] = Field(None, alias="Registrant Organization")
    name: Optional[str] = None
    email: Optional[str] = Field(None, alias="Registrant Email")
    phone: Optional[str] = Field(None, alias="Registrant Phone")
    street: Optional[str] = Field(None, alias="Registrant Street")
    city: Optional[str] = Field(None, alias="Registrant City")
    state: Optional[str] = Field(None, alias="Registrant State/Province")
    country: Optional[str] = None
    postal_code: Optional[str] = Field(None, alias="Registrant Postal Code")
    name_servers: List[str] = Field([], alias="Name Server")
    date_created: Optional[str] = None
    date_changed: Optional[LaxStr] = None
    date_expires: Optional[str] = None
    dnssec: Optional[str] = Field(None, alias="DNSSEC")

    @model_validator(mode="before")
    @classmethod
    def get_latest_whois_record_and_transform(cls, v):
        data = v.pop("data")
        if not data:
            return {}
        transformed = data[0].get("attributes", {})

        # Flatten
        transformed = {**transformed.pop("whois_map"), **transformed}

        # Normalized fields with multiple possible sources
        fields_w_multiple_possible_sources = {
            "date_changed": (
                transformed.pop("last_updated", None)
                or transformed.pop("Updated Date", None)
                or transformed.pop("Last updated", None)
            ),
            "date_created": (
                transformed.pop("Creation Date", None)
                or transformed.pop("Registered on", None)
            ),
            "date_expires": (
                transformed.pop("Expiry Date", None)
                or transformed.pop("Registry Expiry Date", None)
                or transformed.pop("Expiry date", None)
            ),
            "name": (
                transformed.pop("registrant_name", None)
                or transformed.pop("Registrant Name", None)
            ),
            "country": (
                transformed.pop("registrant_country", None)
                or transformed.pop("Registrant Country", None)
            ),
            "registrar": (
                transformed.pop("registrar_name", None)
                or transformed.pop("Registrar", None)
            ),
        }

        return {**transformed, **fields_w_multiple_possible_sources}

    @field_validator("name_servers", mode="before")
    @classmethod
    def transform_nameservers(cls, v):
        return v.lower().split(" | ")

    @field_validator("domain", mode="before")
    @classmethod
    def lowercase_domain(cls, v):
        return v.lower()

    @field_validator("*")
    @classmethod
    def dedupe_values(cls, v):
        if v is not None and "|" in v:
            return v.split(" | ")[0]
        return v



from requests.exceptions import HTTPError

from wtfis.clients.base import BaseRequestsClient, BaseWhoisClient
from wtfis.models.ip2whois import Whois
from wtfis.utils import refang


class Ip2WhoisClient(BaseRequestsClient, BaseWhoisClient):
    """
    IP2WHOIS client
    """

    baseurl = "https://api.ip2whois.com/v2"

    def __init__(self, api_key: str) -> None:
        super().__init__()
        self.api_key = api_key

    @property
    def name(self) -> str:
        return "IP2Whois"

    def get_whois(self, domain: str) -> Whois:
        params = {
            "key": self.api_key,
            "domain": refang(domain),
        }

        # Let a 404 or invalid domain pass
        try:
            return Whois.model_validate(self._get("/", params))
        except HTTPError as e:
            if e.response.status_code == 404 or (
                e.response.status_code == 400
                and e.response.json().get("error", {})["error_code"] == 10007
            ):
                return Whois.model_validate({})
            raise



from typing import Optional

from wtfis.clients.base import BaseIpEnricherClient, BaseRequestsClient
from wtfis.models.ipwhois import IpWhois, IpWhoisMap


class IpWhoisClient(BaseRequestsClient, BaseIpEnricherClient):
    """
    IPWhois client
    """

    baseurl = "https://ipwho.is"

    @property
    def name(self) -> str:
        return "IPWhois"

    def _get_ipwhois(self, ip: str) -> Optional[IpWhois]:
        result = self._get(f"/{ip}")
        return IpWhois.model_validate(result) if result.get("success") is True else None

    def enrich_ips(self, *ips: str) -> IpWhoisMap:
        map_ = {}
        for ip in ips:
            ipwhois = self._get_ipwhois(ip)
            if ipwhois:
                map_[ipwhois.ip] = ipwhois
        return IpWhoisMap.model_validate(map_)



import json
import os
from pathlib import Path
from unittest.mock import MagicMock, patch

import pytest
from dotenv import load_dotenv
from rich.console import Console
from rich.progress import (
    BarColumn,
    Progress,
    SpinnerColumn,
    TaskProgressColumn,
    TextColumn,
    TimeElapsedColumn,
)

from wtfis.clients.abuseipdb import AbuseIpDbClient
from wtfis.clients.base import requests
from wtfis.clients.greynoise import GreynoiseClient
from wtfis.clients.ip2location import Ip2LocationClient
from wtfis.clients.ip2whois import Ip2WhoisClient
from wtfis.clients.ipinfo import IpInfoClient
from wtfis.clients.ipwhois import IpWhoisClient
from wtfis.clients.shodan import ShodanClient
from wtfis.clients.urlhaus import UrlHausClient
from wtfis.clients.virustotal import VTClient
from wtfis.config import Config, parse_args, parse_env
from wtfis.exceptions import HandlerException, WtfisException
from wtfis.handlers.domain import DomainHandler
from wtfis.handlers.ip import IpAddressHandler
from wtfis.main import fetch_data, generate_entity_handler, generate_view, main
from wtfis.models.virustotal import Domain, IpAddress
from wtfis.ui.view import DomainView, IpAddressView

# pylint: disable=protected-access,redefined-outer-name

POSSIBLE_ENV_VARS = [
    "VT_API_KEY",
    "IP2LOCATION_API_KEY",
    "IP2WHOIS_API_KEY",
    "SHODAN_API_KEY",
    "GREYNOISE_API_KEY",
    "ABUSEIPDB_API_KEY",
    "URLHAUS_API_KEY",
    "WTFIS_DEFAULTS",
    "GEOLOCATION_SERVICE",
]


def unset_env_vars():
    for var in POSSIBLE_ENV_VARS:
        try:
            del os.environ[var]
        except KeyError:
            pass


def fake_load_dotenv(tmp_path, fake_env_vars):
    content = [f"{k} = {v}" for k, v in fake_env_vars.items()]
    path = tmp_path / ".env.wtfis"
    path.write_text("\n".join(content))

    def fake(*_):
        load_dotenv(path)

    return fake


def simulate_progress(console):
    return Progress(
        SpinnerColumn(finished_text="[green]✓"),
        TextColumn("[bold]{task.description}"),
        BarColumn(),
        TaskProgressColumn(),
        TimeElapsedColumn(),
        console=console,
    )


@pytest.fixture()
def fake_load_dotenv_1(tmp_path):
    fake_env_vars = {
        "VT_API_KEY": "foo",
        "IP2LOCATION_API_KEY": "foobar",
        "IP2WHOIS_API_KEY": "alice",
        "SHODAN_API_KEY": "hunter2",
        "GREYNOISE_API_KEY": "upupdowndown",
        "ABUSEIPDB_API_KEY": "dummy",
        "URLHAUS_API_KEY": "eve",
    }
    return fake_load_dotenv(tmp_path, fake_env_vars)


@pytest.fixture()
def fake_load_dotenv_2(tmp_path):
    fake_env_vars = {
        "VT_API_KEY": "foo",
        "SHODAN_API_KEY": "hunter2",
        "WTFIS_DEFAULTS": "-s -1",
    }
    return fake_load_dotenv(tmp_path, fake_env_vars)


@pytest.fixture()
def fake_load_dotenv_3(tmp_path):
    fake_env_vars = {
        "VT_API_KEY": "foo",
        "WTFIS_DEFAULTS": "--no-color",
    }
    return fake_load_dotenv(tmp_path, fake_env_vars)


@pytest.fixture()
def fake_load_dotenv_4(tmp_path):
    fake_env_vars = {
        "VT_API_KEY": "foo",
        "GREYNOISE_API_KEY": "bar",
        "ABUSEIPDB_API_KEY": "dummy",
        "WTFIS_DEFAULTS": "-g -u -a",
    }
    return fake_load_dotenv(tmp_path, fake_env_vars)


@pytest.fixture()
def fake_load_dotenv_vt_whois(tmp_path):
    fake_env_vars = {
        "VT_API_KEY": "foo",
    }
    return fake_load_dotenv(tmp_path, fake_env_vars)


@pytest.fixture()
def fake_load_dotenv_ip2whois(tmp_path):
    fake_env_vars = {
        "VT_API_KEY": "foo",
        "IP2WHOIS_API_KEY": "alice",
    }
    return fake_load_dotenv(tmp_path, fake_env_vars)


@pytest.fixture()
def fake_load_dotenv_all_ok(tmp_path):
    fake_env_vars = {
        "VT_API_KEY": "foo",
        "SHODAN_API_KEY": "hunter2",
        "GREYNOISE_API_KEY": "bar",
        "ABUSEIPDB_API_KEY": "dummy",
        "IP2WHOIS_API_KEY": "alice",
        "URLHAUS_API_KEY": "eve",
        "WTFIS_DEFAULTS": "-A",
    }
    return fake_load_dotenv(tmp_path, fake_env_vars)


@pytest.fixture()
def fake_load_dotenv_all_invalid(tmp_path):
    fake_env_vars = {
        "VT_API_KEY": "foo",
        "GREYNOISE_API_KEY": "bar",
        "WTFIS_DEFAULTS": "--all -g",
    }
    return fake_load_dotenv(tmp_path, fake_env_vars)


@pytest.fixture()
def fake_load_dotenv_geolocation_service_invalid(tmp_path):
    fake_env_vars = {
        "VT_API_KEY": "foo",
        "GEOLOCATION_SERVICE": "invalid_service",
    }
    return fake_load_dotenv(tmp_path, fake_env_vars)


class TestArgs:
    def test_basic(self):
        with patch(
            "sys.argv",
            [
                "main",
                "www.example.com",
            ],
        ):
            args = parse_args()
            assert args.entity == "www.example.com"
            assert args.max_resolutions == 3
            assert args.no_color is False
            assert args.one_column is False
            assert args.use_shodan is False
            assert args.use_greynoise is False
            assert args.use_urlhaus is False
            assert args.use_abuseipdb is False
            assert args.all is False

    def test_display(self):
        with patch(
            "sys.argv",
            [
                "main",
                "www.example.com",
                "-n",
                "-1",
            ],
        ):
            args = parse_args()
            assert args.no_color is True
            assert args.one_column is True

    def test_max_resolutions_ok(self):
        with patch(
            "sys.argv",
            [
                "main",
                "www.example.com",
                "-m",
                "8",
            ],
        ):
            args = parse_args()
            assert args.max_resolutions == 8

    def test_max_resolutions_error_1(self, capsys):
        with pytest.raises(SystemExit) as e:
            with patch(
                "sys.argv",
                [
                    "main",
                    "www.example.com",
                    "-m",
                    "11",
                ],
            ):
                parse_args()

        capture = capsys.readouterr()

        assert capture.err.endswith("error: Maximum --max-resolutions value is 10\n")
        assert e.type is SystemExit
        assert e.value.code == 2

    def test_max_resolutions_error_2(self, capsys):
        with pytest.raises(SystemExit) as e:
            with patch(
                "sys.argv",
                [
                    "main",
                    "1.1.1.1",
                    "-m",
                    "5",
                ],
            ):
                parse_args()

        capture = capsys.readouterr()

        assert capture.err.endswith(
            "error: --max-resolutions is not applicable to IPs\n"
        )
        assert e.type is SystemExit
        assert e.value.code == 2

    def test_shodan_ok(self):
        os.environ["SHODAN_API_KEY"] = "foo"
        with patch(
            "sys.argv",
            [
                "main",
                "www.example.com",
                "-s",
            ],
        ):
            args = parse_args()
            assert args.use_shodan is True
        del os.environ["SHODAN_API_KEY"]

    def test_shodan_error(self, capsys):
        with pytest.raises(SystemExit) as e:
            with patch(
                "sys.argv",
                [
                    "main",
                    "www.example.com",
                    "-s",
                ],
            ):
                parse_args()

        capture = capsys.readouterr()

        assert capture.err.endswith("error: SHODAN_API_KEY is not set\n")
        assert e.type is SystemExit
        assert e.value.code == 2

    def test_greynoise_ok(self):
        os.environ["GREYNOISE_API_KEY"] = "foo"
        with patch(
            "sys.argv",
            [
                "main",
                "www.example.com",
                "-g",
            ],
        ):
            args = parse_args()
            assert args.use_greynoise is True
        del os.environ["GREYNOISE_API_KEY"]

    def test_greynoise_error(self, capsys):
        with pytest.raises(SystemExit) as e:
            with patch(
                "sys.argv",
                [
                    "main",
                    "www.example.com",
                    "-g",
                ],
            ):
                parse_args()

        capture = capsys.readouterr()

        assert capture.err.endswith("error: GREYNOISE_API_KEY is not set\n")
        assert e.type is SystemExit
        assert e.value.code == 2

    def test_urlhaus_ok(self):
        os.environ["URLHAUS_API_KEY"] = "foo"
        with patch(
            "sys.argv",
            [
                "main",
                "www.example.com",
                "-u",
            ],
        ):
            args = parse_args()
            assert args.use_urlhaus is True
        del os.environ["URLHAUS_API_KEY"]

    def test_urlhaus_error(self, capsys):
        with pytest.raises(SystemExit) as e:
            with patch(
                "sys.argv",
                [
                    "main",
                    "www.example.com",
                    "-u",
                ],
            ):
                parse_args()

        capture = capsys.readouterr()

        assert capture.err.endswith("error: URLHAUS_API_KEY is not set\n")
        assert e.type is SystemExit
        assert e.value.code == 2

    def test_abuseipdb_ok(self):
        os.environ["ABUSEIPDB_API_KEY"] = "foo"
        with patch(
            "sys.argv",
            [
                "main",
                "1.1.1.1",
                "-a",
            ],
        ):
            args = parse_args()
            assert args.use_abuseipdb is True
        del os.environ["ABUSEIPDB_API_KEY"]

    def test_abuseipdb_error(self, capsys):
        with pytest.raises(SystemExit) as e:
            with patch(
                "sys.argv",
                [
                    "main",
                    "1.1.1.1",
                    "-a",
                ],
            ):
                parse_args()

        capture = capsys.readouterr()

        assert capture.err.endswith("error: ABUSEIPDB_API_KEY is not set\n")
        assert e.type is SystemExit
        assert e.value.code == 2

    def test_all_ok(self):
        os.environ["ABUSEIPDB_API_KEY"] = "foo"
        with patch(
            "sys.argv",
            [
                "main",
                "1.1.1.1",
                "-A",
            ],
        ):
            args = parse_args()
            assert args.all is True
            assert args.use_abuseipdb is False
        del os.environ["ABUSEIPDB_API_KEY"]

    def test_all_error(self, capsys):
        os.environ["SHODAN_API_KEY"] = "foo"
        with pytest.raises(SystemExit) as e:
            with patch(
                "sys.argv",
                [
                    "main",
                    "1.1.1.1",
                    "-A",
                    "-s",
                ],
            ):
                parse_args()

        capture = capsys.readouterr()

        assert capture.err.endswith(
            "error: --use-* flags are not accepted when the --all/-A flag is set\n"
        )
        assert e.type is SystemExit
        assert e.value.code == 2
        del os.environ["SHODAN_API_KEY"]

    def test_ip2location_ok(self):
        os.environ["IP2LOCATION_API_KEY"] = "foo"
        with patch(
            "sys.argv",
            [
                "main",
                "1.1.1.1",
                "--geolocation-service",
                "ip2location",
            ],
        ):
            args = parse_args()
            assert args.geolocation_service == "ip2location"
        del os.environ["IP2LOCATION_API_KEY"]

    def test_ip2location_error(self, capsys):
        with pytest.raises(SystemExit) as e:
            with patch(
                "sys.argv",
                ["main", "1.1.1.1", "--geolocation-service", "ip2location"],
            ):
                parse_args()

        capture = capsys.readouterr()

        assert capture.err.endswith("error: IP2LOCATION_API_KEY is not set\n")
        assert e.type is SystemExit
        assert e.value.code == 2


class TestEnvs:
    def test_env_file(self, fake_load_dotenv_1):
        with patch("wtfis.config.load_dotenv", fake_load_dotenv_1):
            parse_env()
            assert os.environ["VT_API_KEY"] == "foo"
            assert os.environ["IP2LOCATION_API_KEY"] == "foobar"
            assert os.environ["IP2WHOIS_API_KEY"] == "alice"
            assert os.environ["SHODAN_API_KEY"] == "hunter2"
            assert os.environ["GREYNOISE_API_KEY"] == "upupdowndown"
            assert os.environ["ABUSEIPDB_API_KEY"] == "dummy"
        unset_env_vars()

    @patch("wtfis.config.load_dotenv", MagicMock())
    def test_required_env_vars(self):
        os.environ["VT_API_KEY"] = "foo"
        parse_env()
        unset_env_vars()

    @patch("wtfis.config.load_dotenv", MagicMock())
    @patch("wtfis.config.Path.exists")
    def test_error(self, mock_exists, capsys):
        mock_exists.return_value = False

        with pytest.raises(SystemExit) as e:
            parse_env()

        capture = capsys.readouterr()

        assert capture.err == (
            "Error: Environment variable VT_API_KEY not set\n"
            f"Env file {Path().home() / '.env.wtfis'} was not found either. "
            "Did you forget?\n"
        )
        assert e.type is SystemExit
        assert e.value.code == 1

    @patch("sys.argv", ["main", "1.1.1.1"])
    def test_geolocation_service_invalid(
        self, capsys, fake_load_dotenv_geolocation_service_invalid
    ):
        with patch(
            "wtfis.config.load_dotenv", fake_load_dotenv_geolocation_service_invalid
        ):
            with pytest.raises(SystemExit) as e:
                parse_env()
                parse_args()

        capture = capsys.readouterr()

        assert capture.err.endswith(
            "error: Invalid geolocation service: invalid_service. "
            "Valid services are: ip2location, ipinfo, ipwhois\n"
        )
        assert e.type is SystemExit
        assert e.value.code == 2
        unset_env_vars()


class TestDefaults:
    def test_defaults_1(self, fake_load_dotenv_2):
        with patch("wtfis.config.load_dotenv", fake_load_dotenv_2):
            with patch(
                "sys.argv",
                [
                    "main",
                    "www.example.com",
                ],
            ):
                conf = Config()
                assert conf.entity == "www.example.com"
                assert conf.max_resolutions == 3
                assert conf.no_color is False
                assert conf.one_column is True
                assert isinstance(conf.shodan_client, ShodanClient)
                assert conf.greynoise_client is None
                assert conf.abuseipdb_client is None
                assert conf.urlhaus_client is None
                assert isinstance(conf.ip_geoasn_client, IpWhoisClient)
                assert conf.vt_api_key == "foo"
                assert conf.shodan_api_key == "hunter2"
                assert conf.abuseipdb_api_key == ""
                assert conf.greynoise_api_key == ""
                assert conf.ip2whois_api_key == ""
        unset_env_vars()

    def test_defaults_2(self, fake_load_dotenv_2):
        with patch("wtfis.config.load_dotenv", fake_load_dotenv_2):
            with patch(
                "sys.argv",
                [
                    "main",
                    "www.example.com",
                    "-s",
                ],
            ):
                conf = Config()
                assert conf.entity == "www.example.com"
                assert conf.max_resolutions == 3
                assert conf.no_color is False
                assert conf.one_column is True
                assert conf.shodan_client is None
                assert conf.greynoise_client is None
                assert conf.urlhaus_client is None
                assert isinstance(conf.ip_geoasn_client, IpWhoisClient)
        unset_env_vars()

    def test_defaults_3(self, fake_load_dotenv_3):
        with patch("wtfis.config.load_dotenv", fake_load_dotenv_3):
            with patch(
                "sys.argv",
                [
                    "main",
                    "www.example.com",
                ],
            ):
                conf = Config()
                assert conf.entity == "www.example.com"
                assert conf.max_resolutions == 3
                assert conf.no_color is True
                assert conf.one_column is False
                assert conf.shodan_client is None
                assert conf.greynoise_client is None
                assert conf.urlhaus_client is None
        unset_env_vars()

    def test_defaults_4(self, fake_load_dotenv_4):
        with patch("wtfis.config.load_dotenv", fake_load_dotenv_4):
            with patch(
                "sys.argv",
                [
                    "main",
                    "1.1.1.1",
                    "-u",
                ],
            ):
                conf = Config()
                assert conf.entity == "1.1.1.1"
                assert conf.no_color is False
                assert conf.one_column is False
                assert conf.shodan_client is None
                assert conf.urlhaus_client is None
                assert isinstance(conf.greynoise_client, GreynoiseClient)
                assert isinstance(conf.abuseipdb_client, AbuseIpDbClient)
                assert conf.abuseipdb_api_key == "dummy"
        unset_env_vars()


class TestAllFlag:
    def test_1_ok(self, fake_load_dotenv_1):
        """-A flag in argument"""
        with patch("wtfis.config.load_dotenv", fake_load_dotenv_1):
            with patch(
                "sys.argv",
                [
                    "main",
                    "www.example.com",
                    "--all",
                ],
            ):
                conf = Config()
                assert isinstance(conf.shodan_client, ShodanClient)
                assert isinstance(conf.greynoise_client, GreynoiseClient)
                assert isinstance(conf.abuseipdb_client, AbuseIpDbClient)
                assert isinstance(conf.urlhaus_client, UrlHausClient)
        unset_env_vars()

    def test_2_ok(self, fake_load_dotenv_all_ok):
        """-A flag in config file"""
        with patch("wtfis.config.load_dotenv", fake_load_dotenv_all_ok):
            with patch(
                "sys.argv",
                [
                    "main",
                    "www.example.com",
                ],
            ):
                conf = Config()
                assert isinstance(conf.shodan_client, ShodanClient)
                assert isinstance(conf.greynoise_client, GreynoiseClient)
                assert isinstance(conf.abuseipdb_client, AbuseIpDbClient)
                assert isinstance(conf.urlhaus_client, UrlHausClient)
        unset_env_vars()

    def test_3_error(self, fake_load_dotenv_1, capsys):
        """Invalid arguments"""
        with pytest.raises(SystemExit) as e:
            with patch("wtfis.config.load_dotenv", fake_load_dotenv_1):
                with patch(
                    "sys.argv",
                    [
                        "main",
                        "www.example.com",
                        "-A",
                        "-s",
                    ],
                ):
                    _ = Config()

        capture = capsys.readouterr()

        assert capture.err.endswith(
            "error: --use-* flags are not accepted when the --all/-A flag is set\n"
        )
        assert e.type is SystemExit
        assert e.value.code == 2
        unset_env_vars()

    def test_4_error(self, fake_load_dotenv_all_invalid, capsys):
        """Invalid WTFIS_DEFAULTS options"""
        with pytest.raises(SystemExit) as e:
            with patch("wtfis.config.load_dotenv", fake_load_dotenv_all_invalid):
                with patch(
                    "sys.argv",
                    [
                        "main",
                        "www.example.com",
                    ],
                ):
                    _ = Config()

        capture = capsys.readouterr()

        assert capture.err.endswith(
            "error: --use-* flags are not accepted when the --all/-A flag is set\n"
        )
        assert e.type is SystemExit
        assert e.value.code == 2
        unset_env_vars()


class TestGenEntityHandler:
    """Tests for the generate_entity_handler function"""

    @patch("sys.argv", ["main", "www.example[.]com"])
    def test_handler_domain_1(self, fake_load_dotenv_1):
        """Domain with default params"""
        with patch("wtfis.config.load_dotenv", fake_load_dotenv_1):
            conf = Config()
            console = Console()
            entity = generate_entity_handler(conf, console)
        assert isinstance(entity, DomainHandler)
        assert entity.entity == "www.example.com"
        assert entity.max_resolutions == 3
        assert entity.console == console
        assert isinstance(entity._vt, VTClient)
        assert isinstance(entity._geoasn, IpWhoisClient)
        assert isinstance(entity._whois, Ip2WhoisClient)
        assert entity._shodan is None
        assert entity._greynoise is None
        assert entity._urlhaus is None
        assert entity._abuseipdb is None
        unset_env_vars()

    @patch("sys.argv", ["main", "www.example[.]com", "-s", "-g", "-u", "-m", "5"])
    def test_handler_domain_2(self, fake_load_dotenv_1):
        """Domain with Shodan and Greynoise and non-default max_resolutions"""
        with patch("wtfis.config.load_dotenv", fake_load_dotenv_1):
            conf = Config()
            console = Console()
            entity = generate_entity_handler(conf, console)
        assert entity.max_resolutions == 5
        assert isinstance(entity._geoasn, IpWhoisClient)
        assert isinstance(entity._whois, Ip2WhoisClient)
        assert isinstance(entity._shodan, ShodanClient)
        assert isinstance(entity._greynoise, GreynoiseClient)
        assert isinstance(entity._urlhaus, UrlHausClient)
        unset_env_vars()

    @patch("sys.argv", ["main", "www.example[.]com"])
    def test_handler_domain_3(self, fake_load_dotenv_vt_whois):
        """Domain using default Ip2Whois for whois"""
        with patch("wtfis.config.load_dotenv", fake_load_dotenv_vt_whois):
            conf = Config()
            console = Console()
            entity = generate_entity_handler(conf, console)
        assert isinstance(entity._whois, VTClient)
        unset_env_vars()

    @patch("sys.argv", ["main", "www.example[.]com"])
    def test_handler_domain_4(self, fake_load_dotenv_ip2whois):
        """Domain using default Ip2Whois for whois"""
        with patch("wtfis.config.load_dotenv", fake_load_dotenv_ip2whois):
            conf = Config()
            console = Console()
            entity = generate_entity_handler(conf, console)
        assert isinstance(entity._whois, Ip2WhoisClient)
        unset_env_vars()

    @patch("sys.argv", ["main", "1[.]1[.]1[.]1"])
    def test_handler_ip_1(self, fake_load_dotenv_1):
        """IP with default params"""
        with patch("wtfis.config.load_dotenv", fake_load_dotenv_1):
            conf = Config()
            console = Console()
            entity = generate_entity_handler(conf, console)
        assert isinstance(entity, IpAddressHandler)
        assert entity.entity == "1.1.1.1"
        assert entity.console == console
        assert isinstance(entity._vt, VTClient)
        assert isinstance(entity._geoasn, IpWhoisClient)
        assert isinstance(entity._whois, VTClient)
        assert entity._greynoise is None
        assert entity._urlhaus is None
        unset_env_vars()

    @patch("sys.argv", ["main", "1[.]1[.]1[.]1", "-s", "-g", "-u", "-a"])
    def test_handler_ip_2(self, fake_load_dotenv_1):
        """IP with various options"""
        with patch("wtfis.config.load_dotenv", fake_load_dotenv_1):
            conf = Config()
            console = Console()
            entity = generate_entity_handler(conf, console)
        assert isinstance(entity._geoasn, IpWhoisClient)
        assert isinstance(entity._whois, VTClient)
        assert isinstance(entity._shodan, ShodanClient)
        assert isinstance(entity._greynoise, GreynoiseClient)
        assert isinstance(entity._urlhaus, UrlHausClient)
        assert isinstance(entity._abuseipdb, AbuseIpDbClient)
        unset_env_vars()

    @patch("sys.argv", ["main", "1.1.1.1", "--geolocation-service", "ip2location"])
    def test_handler_ip_3(self, fake_load_dotenv_1):
        """IP with ip2location geolocation service"""
        with patch("wtfis.config.load_dotenv", fake_load_dotenv_1):
            conf = Config()
            console = Console()
            entity = generate_entity_handler(conf, console)
        assert isinstance(entity._geoasn, Ip2LocationClient)
        unset_env_vars()

    @patch("sys.argv", ["main", "1.1.1.1", "--geolocation-service", "ipinfo"])
    def test_handler_ip_4(self, fake_load_dotenv_1):
        """IP with ipinfo geolocation service"""
        with patch("wtfis.config.load_dotenv", fake_load_dotenv_1):
            conf = Config()
            console = Console()
            entity = generate_entity_handler(conf, console)
        assert isinstance(entity._geoasn, IpInfoClient)
        unset_env_vars()


class TestGenView:
    """Tests for the generate_view function"""

    @patch("wtfis.main.DomainView", return_value=MagicMock(spec=DomainView))
    def test_view_domain_1(self, m_domain_view, test_data):
        """Domain view with default params"""
        entity = DomainHandler(
            entity=MagicMock(),
            console=MagicMock(),
            vt_client=MagicMock(),
            ip_geoasn_client=MagicMock(),
            whois_client=MagicMock(),
            shodan_client=MagicMock(),
            greynoise_client=MagicMock(),
            abuseipdb_client=MagicMock(),
            urlhaus_client=MagicMock(),
        )
        entity.vt_info = Domain.model_validate(
            json.loads(test_data("vt_domain_gist.json"))
        )
        entity.whois = MagicMock()
        entity.ip_enrich = MagicMock()
        view = generate_view(MagicMock(), MagicMock(), entity)
        assert isinstance(view, DomainView)

    @patch("wtfis.main.IpAddressView", return_value=MagicMock(spec=IpAddressView))
    def test_view_ip_1(self, m_ip_view, test_data):
        """IP address view with default params"""
        entity = IpAddressHandler(
            entity=MagicMock(),
            console=MagicMock(),
            vt_client=MagicMock(),
            ip_geoasn_client=MagicMock(),
            whois_client=MagicMock(),
            shodan_client=MagicMock(),
            greynoise_client=MagicMock(),
            abuseipdb_client=MagicMock(),
            urlhaus_client=MagicMock(),
        )
        entity.vt_info = IpAddress.model_validate(
            json.loads(test_data("vt_ip_1.1.1.1.json"))
        )
        entity.whois = MagicMock()
        entity.ip_enrich = MagicMock()
        view = generate_view(MagicMock(), MagicMock(), entity)
        assert isinstance(view, IpAddressView)

    def test_view_error(self):
        """IP address view with default params"""
        with pytest.raises(WtfisException):
            generate_view(MagicMock(), MagicMock(), "foobar")


class TestMain:
    @patch("sys.argv", ["main", "www.example.com"])
    @patch("wtfis.main.Console", return_value=Console())
    @patch("wtfis.main.Config")
    @patch("wtfis.main.get_progress")
    @patch("wtfis.main.generate_entity_handler", return_value=MagicMock())
    @patch("wtfis.main.generate_view", return_value=MagicMock())
    def test_main_default(
        self, m_view, m_handler, m_progress, m_conf, m_console, fake_load_dotenv_1
    ):
        """Test all calls with default values"""
        m_progress.return_value = simulate_progress(m_console())
        with patch("wtfis.config.load_dotenv", fake_load_dotenv_1):
            m_conf.return_value = Config()
            main()
        m_handler.assert_called_once_with(m_conf(), m_console())
        m_handler().fetch_data.assert_called_once_with()
        m_handler().print_warnings.assert_called_once_with()
        m_view.assert_called_once_with(m_conf(), m_console(), m_handler())
        m_view().print.assert_called_once_with(one_column=False)
        unset_env_vars()


class TestFetchData:
    """
    Test main.fetch_data()
    """

    @patch.object(requests.Session, "get")
    def test_vt_http_error(self, mock_requests_get, domain_handler, capsys):
        """
        Test a requests HTTPError from the VT client. This also tests the
        common_exception_handler decorator.
        """
        handler = domain_handler()
        mock_resp = requests.models.Response()

        mock_resp.status_code = 401
        mock_requests_get.return_value = mock_resp

        # Thorough test of first _fetch_* method
        with pytest.raises(SystemExit) as e:
            fetch_data(MagicMock(), handler)

        capture = capsys.readouterr()

        assert (
            capture.err == "Error fetching data: 401 Client Error: None for url: None\n"
        )
        assert e.type is SystemExit  # ruff E721
        assert e.value.code == 1

        # Extra: just make sure program exits correctly
        with pytest.raises(HandlerException) as e:
            handler._fetch_vt_resolutions()

    @patch.object(requests.Session, "get")
    def test_vt_validation_error(self, mock_requests_get, domain_handler, capsys):
        """
        Test main.fetch_data().
        Test a pydantic data model ValidationError from the VT client. This also tests
        the common_exception_handler decorator.
        """
        handler = domain_handler()
        mock_resp = requests.models.Response()

        with patch.object(mock_resp, "json") as mock_resp_json:
            mock_resp.status_code = 200
            mock_resp_json.return_value = {"intentionally": "wrong data"}
            mock_requests_get.return_value = mock_resp

            # Thorough test of first _fetch_* method
            with pytest.raises(SystemExit) as e:
                fetch_data(MagicMock(), handler)

            capture = capsys.readouterr()

            assert capture.err.startswith(
                "Data model validation error: 1 validation error for Domain\ndata\n"
                "  Field required [type=missing, input_value={'intentionally': "
                "'wrong data'}, input_type=dict]\n"
            )
            assert e.type is SystemExit
            assert e.value.code == 1

            # Extra: just make sure program exits correctly
            with pytest.raises(HandlerException) as e:
                handler._fetch_vt_resolutions()



"""
Global config
"""

import argparse
import os
from argparse import Namespace
from pathlib import Path
from typing import Optional, Union

from dotenv import load_dotenv

from wtfis.clients.abuseipdb import AbuseIpDbClient
from wtfis.clients.greynoise import GreynoiseClient
from wtfis.clients.ip2location import Ip2LocationClient
from wtfis.clients.ip2whois import Ip2WhoisClient
from wtfis.clients.ipinfo import IpInfoClient
from wtfis.clients.ipwhois import IpWhoisClient
from wtfis.clients.shodan import ShodanClient
from wtfis.clients.types import IpGeoAsnClientType
from wtfis.clients.urlhaus import UrlHausClient
from wtfis.clients.virustotal import VTClient
from wtfis.utils import error_and_exit, is_ip
from wtfis.version import get_version

ABUSEIPDB_API_KEY_VAR = "ABUSEIPDB_API_KEY"
GREYNOISE_API_KEY_VAR = "GREYNOISE_API_KEY"
IP2LOCATION_API_KEY_VAR = "IP2LOCATION_API_KEY"
IP2WHOIS_API_KEY_VAR = "IP2WHOIS_API_KEY"
SHODAN_API_KEY_VAR = "SHODAN_API_KEY"
URLHAUS_API_KEY_VAR = "URLHAUS_API_KEY"
VT_API_KEY_VAR = "VT_API_KEY"
WTFIS_DEFAULTS_VAR = "WTFIS_DEFAULTS"
GEOLOCATION_SERVICE_VAR = "GEOLOCATION_SERVICE"


def parse_env() -> None:
    DEFAULT_ENV_FILE = Path().home() / ".env.wtfis"

    # Load the file
    load_dotenv(DEFAULT_ENV_FILE)

    # Exit if required environment variables don't exist
    for envvar in (VT_API_KEY_VAR,):
        if not os.environ.get(envvar):
            error = f"Error: Environment variable {envvar} not set"
            if not DEFAULT_ENV_FILE.exists():
                error += (
                    f"\nEnv file {DEFAULT_ENV_FILE} was not found either. "
                    "Did you forget?"
                )
            error_and_exit(error)


def parse_args() -> Namespace:
    GEOLOCATION_SERVICES = (
        "ip2location",
        "ipinfo",
        "ipwhois",
    )
    DEFAULT_GEOLOCATION_SERVICE = "ipwhois"
    DEFAULT_MAX_RESOLUTIONS = 3

    parser = argparse.ArgumentParser()
    parser.add_argument("entity", help="Hostname, domain or IP")
    parser.add_argument(
        "-A",
        "--all",
        help="Enable all possible enrichments",
        action="store_true",
    )
    parser.add_argument(
        "-s", "--use-shodan", help="Use Shodan to enrich IPs", action="store_true"
    )
    parser.add_argument(
        "-g", "--use-greynoise", help="Enable Greynoise for IPs", action="store_true"
    )
    parser.add_argument(
        "-a", "--use-abuseipdb", help="Enable AbuseIPDB for IPs", action="store_true"
    )
    parser.add_argument(
        "-u",
        "--use-urlhaus",
        help="Enable URLhaus for IPs and domains",
        action="store_true",
    )
    parser.add_argument(
        "-m",
        "--max-resolutions",
        metavar="N",
        help=(
            "Maximum number of resolutions to show "
            f"(default: {DEFAULT_MAX_RESOLUTIONS})"
        ),
        type=int,
        default=DEFAULT_MAX_RESOLUTIONS,
    )
    parser.add_argument(
        "-n", "--no-color", help="Show output without colors", action="store_true"
    )
    parser.add_argument(
        "-1", "--one-column", help="Display results in one column", action="store_true"
    )
    parser.add_argument(
        "-V",
        "--version",
        help="Print version number",
        action="version",
        version=get_version(),
    )
    parser.add_argument(
        "--geolocation-service",
        choices=GEOLOCATION_SERVICES,
        help=f"Geolocation service to use (default: {DEFAULT_GEOLOCATION_SERVICE})",
    )
    parsed = parser.parse_args()

    # Default overrides
    # If a default is set, then setting a boolean flag as an argument _negates_ the effect.
    # Negation does not apply to non-boolean flags (e.g. --max-resolution)
    for option in os.environ.get(WTFIS_DEFAULTS_VAR, "").split(" "):
        if option in ("-A", "--all"):
            parsed.all = not parsed.all
        elif option in ("-s", "--use-shodan"):
            parsed.use_shodan = not parsed.use_shodan
        elif option in ("-g", "--use-greynoise"):
            parsed.use_greynoise = not parsed.use_greynoise
        elif option in ("-a", "--use-abuseipdb"):
            parsed.use_abuseipdb = not parsed.use_abuseipdb
        elif option in ("-u", "--use-urlhaus"):
            parsed.use_urlhaus = not parsed.use_urlhaus
        elif option in ("-n", "--no-color"):
            parsed.no_color = not parsed.no_color
        elif option in ("-1", "--one-column"):
            parsed.one_column = not parsed.one_column

    # Geolocation service
    # Commandline flag takes precedence over env file or variable
    # Set the default here if no service is provided
    parsed.geolocation_service = (
        parsed.geolocation_service
        or os.environ.get(GEOLOCATION_SERVICE_VAR)
        or DEFAULT_GEOLOCATION_SERVICE
    )

    # Validations
    if parsed.max_resolutions > 10:
        argparse.ArgumentParser().error("Maximum --max-resolutions value is 10")
    if is_ip(parsed.entity) and parsed.max_resolutions != DEFAULT_MAX_RESOLUTIONS:
        argparse.ArgumentParser().error("--max-resolutions is not applicable to IPs")

    if parsed.use_shodan and not os.environ.get(SHODAN_API_KEY_VAR):
        argparse.ArgumentParser().error(f"{SHODAN_API_KEY_VAR} is not set")
    if parsed.use_greynoise and not os.environ.get(GREYNOISE_API_KEY_VAR):
        argparse.ArgumentParser().error(f"{GREYNOISE_API_KEY_VAR} is not set")
    if parsed.use_abuseipdb and not os.environ.get(ABUSEIPDB_API_KEY_VAR):
        argparse.ArgumentParser().error(f"{ABUSEIPDB_API_KEY_VAR} is not set")
    if parsed.use_urlhaus and not os.environ.get(URLHAUS_API_KEY_VAR):
        argparse.ArgumentParser().error(f"{URLHAUS_API_KEY_VAR} is not set")

    if parsed.geolocation_service not in GEOLOCATION_SERVICES:
        argparse.ArgumentParser().error(
            f"Invalid geolocation service: {parsed.geolocation_service}. "
            f"Valid services are: {', '.join(GEOLOCATION_SERVICES)}"
        )
    if parsed.geolocation_service == "ip2location" and not os.environ.get(
        IP2LOCATION_API_KEY_VAR
    ):
        argparse.ArgumentParser().error(f"{IP2LOCATION_API_KEY_VAR} is not set")

    if parsed.all and (
        parsed.use_shodan
        or parsed.use_greynoise
        or parsed.use_abuseipdb
        or parsed.use_urlhaus
    ):
        argparse.ArgumentParser().error(
            "--use-* flags are not accepted when the " "--all/-A flag is set"
        )

    return parsed


class Config:
    def __init__(self) -> None:
        # Load environment variables
        parse_env()

        # Get arg values
        self.args: Namespace = parse_args()

        # Creds
        self.abuseipdb_api_key = os.environ.get(ABUSEIPDB_API_KEY_VAR, "")
        self.greynoise_api_key = os.environ.get(GREYNOISE_API_KEY_VAR, "")
        self.ip2location_api_key = os.environ.get(IP2LOCATION_API_KEY_VAR, "")
        self.ip2whois_api_key = os.environ.get(IP2WHOIS_API_KEY_VAR, "")
        self.shodan_api_key = os.environ.get(SHODAN_API_KEY_VAR, "")
        self.urlhaus_api_key = os.environ.get(URLHAUS_API_KEY_VAR, "")
        self.vt_api_key = os.environ.get(VT_API_KEY_VAR, "")

    @property
    def entity(self) -> str:
        return self.args.entity

    @property
    def max_resolutions(self) -> int:
        return self.args.max_resolutions

    @property
    def no_color(self) -> bool:
        return self.args.no_color

    @property
    def one_column(self) -> bool:
        return self.args.one_column

    @property
    def vt_client(self) -> VTClient:
        return VTClient(self.vt_api_key)

    @property
    def ip_geoasn_client(self) -> IpGeoAsnClientType:
        # IP geolocation and ASN client selector
        if self.args.geolocation_service == "ip2location":
            return Ip2LocationClient(self.ip2location_api_key)
        if self.args.geolocation_service == "ipinfo":
            return IpInfoClient()
        return IpWhoisClient()

    @property
    def whois_client(self) -> Union[Ip2WhoisClient, VTClient]:
        # Whois client selector
        # Order of use based on set envvars:
        #    1. IP2Whois (Domain only)
        #    2. Virustotal (fallback)
        if self.ip2whois_api_key and not is_ip(self.entity):
            return Ip2WhoisClient(self.ip2whois_api_key)
        return self.vt_client

    @property
    def abuseipdb_client(self) -> Optional[AbuseIpDbClient]:
        if (self.args.use_abuseipdb or self.args.all) and bool(self.abuseipdb_api_key):
            return AbuseIpDbClient(self.abuseipdb_api_key)
        return None

    @property
    def greynoise_client(self) -> Optional[GreynoiseClient]:
        if (self.args.use_greynoise or self.args.all) and bool(self.greynoise_api_key):
            return GreynoiseClient(self.greynoise_api_key)
        return None

    @property
    def shodan_client(self) -> Optional[ShodanClient]:
        if (self.args.use_shodan or self.args.all) and bool(self.shodan_api_key):
            return ShodanClient(self.shodan_api_key)
        return None

    @property
    def urlhaus_client(self) -> Optional[UrlHausClient]:
        if self.args.use_urlhaus or self.args.all and bool(self.urlhaus_api_key):
            return UrlHausClient(self.urlhaus_api_key)
        return None



from __future__ import annotations

from typing import Dict, Optional

from pydantic import BaseModel, RootModel


class GreynoiseIp(BaseModel):
    ip: str
    noise: bool
    riot: bool
    message: str
    link: str
    classification: Optional[str] = None
    name: Optional[str] = None
    last_seen: Optional[str] = None


class GreynoiseIpMap(RootModel):
    root: Dict[str, GreynoiseIp]

    @classmethod
    def empty(cls) -> GreynoiseIpMap:
        return cls.model_validate({})



"""
Logic handler for IP address inputs
"""

from wtfis.handlers.base import (
    BaseHandler,
    common_exception_handler,
    failopen_exception_handler,
)


class IpAddressHandler(BaseHandler):
    @common_exception_handler
    def _fetch_vt_ip_address(self) -> None:
        self.vt_info = self._vt.get_ip_address(self.entity)

    @common_exception_handler
    @failopen_exception_handler("_urlhaus")
    def _fetch_urlhaus(self) -> None:
        if self._urlhaus:
            self.urlhaus = self._urlhaus.enrich_ips(self.entity)

    def fetch_data(self):
        yield "Fetching data from Virustotal", 50
        self._fetch_vt_ip_address()

        yield f"Fetching IP location and ASN from {self._geoasn.name}", 50
        self._fetch_geoasn(self.entity)

        if self._shodan:
            yield f"Fetching IP data from {self._shodan.name}", 50
            self._fetch_shodan(self.entity)

        if self._urlhaus:
            yield f"Fetching IP data from {self._urlhaus.name}", 50
            self._fetch_urlhaus()

        if self._greynoise:
            yield f"Fetching IP data from {self._greynoise.name}", 50
            self._fetch_greynoise(self.entity)

        if self._abuseipdb:
            yield f"Fetching IP data from {self._abuseipdb.name}", 50
            self._fetch_abuseipdb(self.entity)

        yield f"Fetching IP whois from {self._whois.name}", 50
        self._fetch_whois()
